//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FLogWindow.h - Implements a simple logfile window using a QListView
//
// $Id: FLogWindow.C,v 1.14 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FLogWindow.moc"
#include "FLogWindow.h"

#include <stdio.h>

FLogWindow::FLogWindow(QWidget * parent, const char * name )
  : QListView( parent, name ), maxEntryIcons(FColors::numAgentColors())
{
  entryIcons= new (QPixmap*)[FColors::numAgentColors()+1];
  for (int i=-1; i < FColors::numAgentColors(); i++)
    entryIcons[i+1]= FColors::createColorIcon(i);

  addColumn("Time");
  addColumn("Logfile entry");

  contextMenu= new QPopupMenu;
  contextMenu->insertItem("Clear", this, SLOT(clearList()), 0, 0);
  contextMenu->insertItem("Only one step", this, SLOT(toggleClearMode()),
			  0, 1);
  contextMenu->setItemChecked(1, true);
  setCaption("FUNSSELMon LogWindow");
  resize(Config::getValueAsInt("FMonLogSizeX"),
	 Config::getValueAsInt("FMonLogSizeY"));
}

FLogWindow::~FLogWindow()
{
  delete [] entryIcons;
}

void FLogWindow::addLine( const int time, const int color, 
			  const int plNum, const char *msg)
{
  char buf[20] = "";
  int col( color>-1 ? color < maxEntryIcons ? color+1 : 0 : 0 );
  sprintf(buf,"%5.5d", time);
  QListViewItem *lvi= new QListViewItem( this );
  lvi->setText(0,buf);
  lvi->setPixmap(1,*(entryIcons[col]));
  char *buf2= new char[strlen(msg)+10];  
  if (plNum > -1) {
    sprintf(buf2, "%d ", plNum);
  } else {
    strcpy(buf2, "");
  }
  strncat(buf2, msg, strlen(msg)-1);
  lvi->setText(1, buf2);
}

#ifdef QT2
void FLogWindow::contentsMouseReleaseEvent ( QMouseEvent * e )
#else
void FLogWindow::mouseReleaseEvent ( QMouseEvent * e )
#endif
{
  if (e->button() == RightButton)
    contextMenu->popup(e->globalPos());

  QListView::mouseReleaseEvent( e );
}

