//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef FILTER_SOCKET_H
#define FILTER_SCOKET_H
#include <string>
#include <list>
#include <arpa/inet.h>
#include "Exception.h"

class FilterMessage
{
public:
  string   message;
  int      client;
  bool     fromSvr;

public:
  enum {
    ToServer   = false,
    FromServer = true
  };

  FilterMessage(const string &msg, int clnt, bool d)
    : message(msg), client(clnt), fromSvr(d) {}
  FilterMessage(const FilterMessage &m): message(m.message), client(m.client),
					 fromSvr(m.fromSvr) {}

  bool fromClient()  const { return !fromSvr; }
  bool fromServer()  const { return fromSvr;  }
  bool isValid()     const { return (client >= 0); };
};

/**
   This class provides an proxy interface for socket like the ones used in the
   soccerserver. Data sent from the soccerserver and the clients
   (players, monitors etc) is caught and reported via the getNextMessage()
   callback. Messages are sent to the server or a client via send().
   if you send() all messages caught by getNextMessage(), this class acts as a
   sniffer.
*/
class FilterSocket
{
public:
  FilterSocket(int fakePort, const char *remote_host,
	       int sniffedPort, int maxMessageSize)
    throw (Exception);
  ~FilterSocket() throw (Exception);

  /// enable SIGIO to get interrupted when new data is avaiable
  static void enableInts();

  /** get next message from communication with soccerserver.
      Returns all messages sent from or to server.
   */
  FilterMessage getNextMessage() throw (Exception);

  /** drop a client.
   */
  void drop(int clientID) throw (Exception);

  /** connect directly to server.
      This method creates a new connection to the soccerserver
      which is not associated with any client.
      You can't send data to this client, but you can receive data
      from the server or send data to the server.
      Returns the associated ID.
  */
  int connect() throw (Exception);

  /** send message for or to a client.
      If the message has the fromServer flag set, the message is sent to the
      client as if the soccerserver would have sent it,
      else the soccerserver receives the message "from" the client
  */
  void send(const FilterMessage &msg) throw (Exception);

  /** find out if a client exists */
  bool clientExists(int client);
private:
  class FilterClient
  {
    friend class FilterSocket;

    sockaddr_in addr;
    sockaddr_in mysrv;
    int sock;
  };

  void sendToClient(const FilterClient &client, const char *buf, int n) throw (Exception);
  void sendForClient(const FilterClient &client, const char *buf, int n) throw (Exception);

  const FilterClient &findClient(sockaddr_in addr);

  static void myhandler(int);
  
  FilterClient       primary;
  list<FilterClient> clients;

  int         maxMsgSize;
};

#endif
