//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include <constants.h>
#include "Recoder.h"
#include <Config.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <errno.h>
#include <netinet/in.h>

#define FORCE_CORE
//#define DEBUG

// cwz-build-command: make recoder

Recoder::Recoder(const string &_logfile)
  throw (Exception) :
  messages(_logfile),
  log_board(!Config::getValueAsInt("filter_board_log")),
  msg_board(!Config::getValueAsInt("filter_board_audio")),
  time(0)
{
  messages.openWrite();
}

Recoder::~Recoder()
{
}

///////////////////////////////////////// code

template<class c>
void iread(c &a)
{
  cin.read((char *)&a, sizeof(c));
}

void Recoder::code() throw (Exception)
{
  char buf[256];
  cin.read(buf, 4);
  if (!((buf[0] == 'U') && (buf[1] == 'L') &&
	(buf[2] == 'G') && (buf[3] == REC_VERSION_3)))
    throw Exception("can only convert v3 files");
  
  cerr << "converting" << endl;

  while (cin.good()) {
    short       len;

    iread(m_tBuf2.mode);
    switch (ntohs(m_tBuf2.mode)) {
    case PM_MODE:
      iread(m_playmode_buf);
      break;
    case TEAM_MODE:
      iread(m_team_buf[0]);
      iread(m_team_buf[1]);
      break;
    case SHOW_MODE: {
      m_tBuf2.body.show.pmode = m_playmode_buf;
      m_tBuf2.body.show.team[0] = m_team_buf[0];
      m_tBuf2.body.show.team[1] = m_team_buf[1];
			
      // read a show struct
      iread(m_show_buf);
      m_tBuf2.body.show.ball = m_show_buf.ball;
      for (int i = 0; i < 2*MAX_PLAYER; i++)
	m_tBuf2.body.show.pos[i] = m_show_buf.pos[i];
      m_tBuf2.body.show.time = m_show_buf.time;

      // got a full SHOW_MODE, store away
      // update store time
      if (ntohs(m_show_buf.time) != time) {
	if (ntohs(m_show_buf.time) != time + 1)
	  cerr << "Warning: time skew: last=" << time << " curr=" << ntohs(m_show_buf.time) << endl;
	cerr << "\r" << "time " << time;
	time = ntohs(m_show_buf.time);
      }

      SoccerMessage s_msg(time, "", SoccerServer, FunsselMonitor);
      s_msg.message.assign((char *)&m_tBuf2, sizeof(Soccer::dispinfo_t2));
      messages.insertMsg(s_msg);
      break;
    }
    case MSG_MODE: {
      // read the board info
      iread(m_tBuf2.body.msg.board);

      // read the string length
      iread(len);
      len = ntohs(len);
		
      // read the message
      cin.read(m_tBuf2.body.msg.message, len);

      // got a full MSG_MODE, store away
      SoccerMessage s_msg(time, "", SoccerServer, FunsselMonitor);
      s_msg.message.assign((char *)&m_tBuf2, sizeof(Soccer::dispinfo_t2));
      messages.insertMsg(s_msg);
      break;
    }
    case PARAM_MODE:
      iread(m_tBuf2.body.sparams);
      break;
    case PPARAM_MODE:
      iread(m_tBuf2.body.pparams);
      break;
    case PT_MODE:
      iread(m_tBuf2.body.ptinfo);
      break;
    }
  }
  cerr << endl;
}

int main(int argc, char **argv)
{
  Config::readFile("funsselrc");

  string logfile  = "soccerlog";

  char c;
  while ((c = getopt(argc, argv, "sl:")) != EOF) {
    switch (c) {
    case 'l':
      logfile  = optarg;
      break;
    default: ;
    }
  }

#ifdef FORCE_CORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;  
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
  
  try {
    Recoder recoder(logfile);
    recoder.code();
  }
  catch (Exception e) {
    cerr << "an exception occured: " << e << endl;
  }
  catch (...) {
    cerr << "an unexpected exception occured" << endl;
  }
  return 0;
}
