//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Socket.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <iomanip>
#include <signal.h>

//#define DEBUG
//#define TRANSMIT

class Socket_Error: public Exception {
public:
  Socket_Error(const char *msg, int errno_val);
};

Socket_Error::Socket_Error(const char *error, int err)
{
  char buf[256];
  sprintf(buf, "socket: errno=%d during %s", err, error);
  descr = buf;
}

Socket::Socket(int port, const char *remote_host,
	       int maxSize)
  throw (Exception): maxMsgSize(maxSize)
{
  sock = socket(AF_INET, SOCK_DGRAM, 0);
  fcntl(sock, F_SETFL, O_NONBLOCK|O_ASYNC);
  fcntl(sock, F_SETOWN, getpid());

  // construct server in_addr
  bzero((char *)&srv, sizeof(srv));
  srv.sin_family      = AF_INET;
  struct hostent *hp;
  hp = gethostbyname(remote_host);
  if (!hp)
    throw Socket_Error("server host unknown", -1);
  srv.sin_addr = *((in_addr *)hp->h_addr);
  srv.sin_port = htons(port);

  // bind socket to a port
  struct sockaddr_in saddr;
  bzero((char *)&saddr, sizeof(saddr));
  saddr.sin_family      = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  saddr.sin_port        = htons(0);
  if (bind(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    throw Socket_Error("bind socket", errno);
}

Socket::~Socket() throw (Exception)
{
  if (close(sock) < 0)
    throw Socket_Error("close port", errno);
}

string Socket::poll() throw (Exception)
{
  sockaddr_in caddr;
  size_t size = sizeof(caddr);
  char buf[maxMsgSize];
  int  n = recvfrom(sock, (void *)buf, maxMsgSize,
		    0, (sockaddr *)&caddr, &size);
  if ((n == -1) && (errno != EAGAIN))
    throw Socket_Error("polling port", errno);
  if (n > 0) {
    string msg(buf, n);
#ifdef DEBUG
    cout << "recvd from " << ntohs(client.addr.sin_port)
	 << " " << hex << client.addr.sin_addr.s_addr << dec << endl;
#ifdef DEBUG_TRANSMIT
    if (n < 512) {
      char test[n + 1];
      memcpy(test, buf, n);
      test[n] = 0;
      cout << "<<" << test << ">>" << n << endl;
    }
    else
      cout << "msg len " << n << endl;
#endif
#endif
    // now NODA special: send to from where you receive
    srv = caddr;
    return msg;
  }
  return string();
}

void Socket::send(const string &msg) throw (Exception)
{
  int n = msg.length();
#ifdef DEBUG
  cout << "transmission to " << ntohs(client.addr.sin_port)
       << " " << hex << client.addr.sin_addr.s_addr << dec << endl;
#ifdef DEBUG_TRANSMIT
  if (n < 512) {
    char test[n + 1];
    memcpy(test, msg.data(), n);
    test[n] = 0;
    cout << "<<" << test << ">>" << n << endl;
  }
  else
    cout << "msg len " << n << endl;
#endif
#endif
  int sent = sendto(sock, (void *)msg.data(), n,
		    0, (sockaddr *)&srv, sizeof(srv));
  if (n != sent)
    throw Socket_Error("transmission", errno);
}

void Socket::myhandler(int)
{
}


void Socket::enableInts()
{
  sigset_t sset;
  sigemptyset(&sset);
  struct sigaction saction;
  saction.sa_handler = myhandler;
  saction.sa_flags   = SA_RESTART;
  saction.sa_mask    = sset;
  sigaddset(&sset, SIGIO);
  sigprocmask(SIG_UNBLOCK, &sset, NULL);
  sigaction(SIGIO, &saction, NULL);
}
