//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: audiomgr.C,v 1.14 2000/10/19 15:27:43 goetz Exp $
//
//////////////////////////////////////////////////////////////////////

#include "audiomgr.h"

#include <iostream.h>
#include <stdio.h>
#include <string.h>

#include "constants.h"
#include "coordinates.h"

#include "snapshot.h"
#include "audiomsg.h"

//////////////////////////////////////////////////////////////////////
//
// class AudioMsgQueue
//
//////////////////////////////////////////////////////////////////////

AudioMsgQueue::AudioMsgQueue( const unsigned int no) :
  queueLen(no), readMsg(0), writeMsg(0), numMessages(0)
{
  queue= new AudioMsg[no];
}
  
AudioMsgQueue::~AudioMsgQueue()
{
  delete [] queue;
}

bool AudioMsgQueue::getNextMsg(AudioMsg &msg)
{
  if ( numMessages > 0 ){
    msg = queue[readMsg];
    readMsg = ( readMsg+1 ) % queueLen;
    numMessages--;
    return true;
  } else {
    // Queue is empty
    return false;
  }
}

void AudioMsgQueue::storeMessage(const AudioMsg &msg)
{
  // Adjust read-pointer if overflow of queue
  if ( ( numMessages > 0 ) && ( writeMsg == readMsg ) ){
    readMsg= ( readMsg+1 ) % queueLen;
  }

  // Add message to queue
  queue[writeMsg]= AudioMsg(msg);
  writeMsg= ( writeMsg+1 ) % queueLen;
  if ( numMessages < queueLen ){
    numMessages++;
  }
}

//////////////////////////////////////////////////////////////////////
//
// class AudioMgr
//
//////////////////////////////////////////////////////////////////////

AudioMgr::AudioMgr(const unsigned int myPlayerNum, 
		   const bool sideIsLeft,
		   const char msgPrefix[4],
		   const unsigned int queueSize) :
  myNumber(myPlayerNum), mySideIsLeft(sideIsLeft)
{
  ensure( queueSize > 0);
  strcpy(m_prefix, msgPrefix);
  myQueue= new AudioMsgQueue( queueSize );
}

AudioMgr::~AudioMgr()
{
  delete myQueue;
}

void AudioMgr::setStates(int number, bool sideIsLeft) {
  myNumber= number;
  mySideIsLeft= sideIsLeft;
}

const char* AudioMgr::SpeakTo( AudioMsg &msg)
{
  msg.setPrefix(m_prefix);
  msg.setSourceAndSideIsLeft(myNumber, mySideIsLeft);
  msg.rewindRead().createChkSums();
  
  return msg.getEncoding();
}

bool AudioMgr::getNextMsg( AudioMsg &msg)
{
  return myQueue->getNextMsg( msg );
}

bool AudioMgr::storeMsg(const AudioMsg &msg )
{
  if (!msg.isChkSumsValid() || 
      !msg.verifyPrefix(m_prefix) ||
      (mySideIsLeft != msg.getSideIsLeft()))
    return false;
  if (!msg.verifyDestination(myNumber)) 
    return false;
  timeOfLastHeard= msg.getServerTime();
  myQueue->storeMessage( msg );
  return true;
}

bool AudioMgr::heardMsgRecently(unsigned int stime)
{
  return timeOfLastHeard < stime - 1;
}
