//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: comm.C,v 1.47 2001/06/26 17:29:08 satan Exp $
//  $Manager: eri$
// 
// 
//  comm.C - Communication with soccer server
// 
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string>
#include <iostream.h>
#include <sys/file.h>
#include <sys/resource.h>

#include <fcntl.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "comm.h"
#include "FunsselStrings.h"
//
// Remember: slow_down_factor (since sserver-6.06) has not been implemented yet
//

///////////////////////////////////////////////////////////////////////

#undef CERR
#define CERR(output) cerr << output << endl

//////////////////////////////////////////////////////////////////////

#undef INITTRACE
#undef SERVERTRACE
#undef TRACE
#undef SIGNALTRACE
#undef SAYTRACE

/* uncomment #define... to activate trace */

// prompt initialization process
//#define INITTRACE(cmd) cmd

// prompt all communication to and from server
//#define SERVERTRACE(cmd) cmd

// prompt all params sent from server
//#define PARAMTRACE(text,val) cerr << ##text## << ##val## << endl;

// general prompting of work
//#define TRACE(cmd) cmd

// prompt incoming signals (careful - just for debugging purposes !! )
// ( Enabling this will almost surely cause segfaults during a normal game.
//   It will also mess up the output stream. )
//#define SIGNALTRACE(cmd) cmd

// prompt encoding/decoding of say messages
//#define SAYTRACE(cmd) cmd

#ifndef INITTRACE
#define INITTRACE(cmd)
#else
#warning *** Comm: INITTRACE macro enabled
#endif

#ifndef SERVERTRACE
#define SERVERTRACE(cmd)
#else
#warning *** Comm: SERVERTRACE macro enabled
#endif

#ifndef PARAMTRACE
#define PARAMTRACE(cmd1, cmd2)
#else
#warning *** Comm: PARAMTRACE macro enabled
#endif

#ifndef TRACE
#define TRACE(cmd)
#else
#warning *** Comm: TRACE macro enabled
#endif

#ifndef SIGNALTRACE
#define SIGNALTRACE(cmd)
#else
#warning *** Comm: SIGNALTRACE macro enabled
#endif

#ifndef SAYTRACE
#define SAYTRACE(cmd)
#else
#warning *** Comm: COACHTRACE macro enabled
#endif

#ifndef CERR
#define CERR(output)
#warning *** Comm error output macro CERR disabled!
#endif

////////////////////////////////////////////////////////////////////


// define this symbol to use pause instead of sigsuspend 
// (pause is more reliable at the moment, 
//  sigsuspend is the more elegant solution)
#undef USE_PAUSE
#define USE_PAUSE


/////////////////////////////////////////////////////////////////////


// initialize the static Comm pointer once at the beginning, to define it.
Comm *(Comm::theComm) = NULL;

// static member methods for signal handling
void Comm::alarmHandler(int signo) {

  if (signo) {} // prevent warning when compiling
  if (Comm::theComm != NULL) {
    SIGNALTRACE( cout << "SIGALRM" << endl);
    Comm::theComm->alarmAction(); 
  } else {
    CERR("Comm: cannot handle SIGALRM as Comm is missing !" );
  }
  
}

void Comm::ioHandler(int signo) {

  if (signo) {} // prevent warning when compiling
  if (Comm::theComm != NULL) {
    SIGNALTRACE( cout << "SIGIO" << endl);
    Comm::theComm->ioAction(); 
  } else {
    CERR("Comm: cannot handle SIGIO as Comm is missing !" );
  }

}
  

///////////////////////////////////////////////////////////////////////////


Comm::Comm(const char* _team, const bool FUNSSELmode) 
  : FUNSSELmode(FUNSSELmode)
{
  TRACE( cout << "Comm: constructor entered" << endl);

  if (Comm::theComm == NULL) { // only if no other Comm exists yet
    Comm::theComm = this; // store static reference

    sd = -1;
    serverOk = false;

    sigemptyset(&iomask);
    sigaddset(&iomask, SIGIO);
    sigemptyset(&alrmask);
    sigaddset(&alrmask, SIGALRM);
    alarmFlag = 0;
    ioFlag    = 0;
    
    // [HACK]
    info = new AsyncCommInfo(_team, this);
    parser   = new Parser();
    store    = new SensorDataStore();
    audiomgr = new AudioMgr();

    hadMessage = false;
    noMessageCount = 0;
    printCommands  = false;
    
  } else {
    CERR("Comm: cannot construct more than one Comm "
	 << "in any single process !");
  }

  TRACE( cout << "Comm: constructor completed" << endl);

#ifndef NOCORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;  
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
}


Comm::~Comm() {

  TRACE( cout << "Comm: destructor entered" << endl);

  // log out
  if (serverOk) sendToServer ("(bye)");

  removeSigHandlers();
  
  // remove static reference
  Comm::theComm = NULL;
  
  delete info;
  if (sd >= 0) close(sd);
  if (parser) delete parser;
  if (store) delete store;
  if (audiomgr) delete audiomgr;

  TRACE( cout << "Comm: destructor completed" << endl);

}

int
Comm::getTimeExpression() {
  const int MAX_SEC = 100;
  struct timeval now;
  gettimeofday (&now, NULL);
  long time_ms = (now.tv_sec%MAX_SEC)*1000 + now.tv_usec/1000;
  return time_ms % 10000;
}

comm_error 
Comm::sendToServer(const char *msg) {

  // Prints the string and sends it to the server via the UDP socket.
      
  SERVERTRACE( cout << getServerTime() << " systime: " << getTimeExpression() 
	       << " snd: "<< msg << endl);
  
  if (printCommands) {
    if (strncmp(msg, "(say", 4) != 0) { // avoid logging say commands
      cout << info->getServerTime() << "> " << msg << endl;
    }
  }
  
  int len = strlen(msg);
  int actual;
  do { // repeat sending if interrupted
    actual = sendto(sd, msg, len, 0, (struct sockaddr *)
		    &server_addr, sizeof(server_addr));
  } while (actual < 0 && errno == EINTR);
  
  if (actual != len) return COMM_ERROR;
  else return COMM_OK;
}


comm_error 
Comm::receiveFromServer(char *msg, long timeout_msec) {
  // select() the socket to find out if data is there. Then keep receiving
  // until the whole string was read from the socket, and return it.
  
  fd_set                watchMe;
  timeval               timeout;
  int	                n;

#ifdef Linux
  unsigned int          fromlen;
#else
  int                   fromlen;
#endif

  struct sockaddr_in    from_addr;
  bool                  repeatSelect;
  
  // create the file descriptor set to be watched by select()
  FD_ZERO (&watchMe);
  FD_SET (sd, &watchMe);
  timeout.tv_sec  = (1000*timeout_msec) / 1000000;
  timeout.tv_usec = (1000*timeout_msec) % 1000000;
  
  do {
    repeatSelect = false;
    switch (select (sd+1, &watchMe, NULL, NULL, &timeout)) {
      
    case 1: // the socket has data
      fromlen = sizeof(from_addr) ;
      
      do { // repeat receiving if interrupted
	n = recvfrom(sd, msg, MAX_SERVER_MSG_SIZE, 0,
		     (struct sockaddr *)&from_addr, &fromlen);
      } 
      while ((n == -1) && (errno == EINTR));
      
      if (n<0) return COMM_ERROR; // not an EINTR error !
      msg[n] = 0;

      SERVERTRACE(cout << getServerTime() << " systime: " << getTimeExpression() 
		  << " rcv: " << msg << endl);

      // sender of this message will be next adressee (when receiving init)
      server_addr.sin_port = from_addr.sin_port;

      return COMM_OK; 
      
    case 0: // socket had no data and timed out
      return COMM_ERROR_NO_DATA;
      
    case -1: // select() returned an error
      if (errno != EINTR) return COMM_ERROR; // select returned a serious error
      else repeatSelect = true; // repeat on interrupt
    } //switch
  } //do
  while (repeatSelect);
    
  // if control would reach this point, something were seriously wrong
  return COMM_ERROR;
  
}

void 
Comm::alarmAction(){

  alarmFlag = 1;
  
}
  
void
Comm::finishAlarm() {
  
  TRACE(cout << "Comm: handling SIGALRM" << endl);
  
  if (hadMessage) {
    noMessageCount = 0;
    hadMessage = false;
  } else {
    TRACE(cout << "Comm: no message from server since last alarm (for " 
	  << noMessageCount+1 << " steps now)" << endl);
    noMessageCount++;
  }
  
  if (noMessageCount > COMM_MAX_DEAD_SERVER_CYCLES) {
    TRACE(cout << "Comm: no messages for more than " 
	  << COMM_MAX_DEAD_SERVER_CYCLES 
	  << " supposed timesteps, server is dead." << endl);
    serverOk = false;
  }
  
}

void 
Comm::ioAction(){

  ioFlag = 1;

}

void 
Comm::finishIo() {

  TRACE(cout << "Comm: handling SIGIO" << endl);
  char msg [MAX_SERVER_MSG_SIZE];
  
  while (receiveFromServer(msg, 0) == COMM_OK) {
    
    hadMessage = true;
    
    SensorData data;
    if (!parser->parseServerString(msg, data)) {
      CERR(getServerTime() 
	   << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): could not parse server string " << msg);
    } else {
      
      // use data to adjust time, play mode etc.
      bool firstSenseBody = info->adjust(data); 

      switch (data.type) {

      case SENSOR_TYPE_OK:
	TRACE(cout << "Comm: skipping ok message" << endl);
	break;
      case SENSOR_TYPE_ERROR: 
	CERR(getServerTime()
	     << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): received error message from server: " << msg );
	break;
	
      case SENSOR_TYPE_SENSE_BODY:
	if (info->isTimeRunning() && firstSenseBody) {
	  // first sense_body of a timestep received while clock is running:
	  // insert it before all data from same timestep 
	  // (needed by worldmodel)
	  TRACE(cout << "Comm: inserting first sense_body data into store" 
		<< endl);
	  store->insert(data);
	} else {
	  // additional sense_body, or sense_body received while clock 
	  // is stopped: just store it according to receive time
	  TRACE(cout << "Comm: storing sense_body data" << endl);
	  store->put(data);
	}
	break;
      case SENSOR_TYPE_EXACT:
	TRACE(cout << "Comm: storing exact data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_SEE:
	TRACE(cout << "Comm: storing see data" << endl);
	store->put(data); // worldmodel can handle twisted neck
	break;
      case SENSOR_TYPE_PLAYER_TYPE:
	TRACE(cout << "Comm: player_type" << endl);
	TRACE(cout << "Comm: storing player_type data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_PLAYER_CHANGED:
	TRACE(cout << "Comm: player changed" << endl);
	TRACE(cout << "Comm: storing player_changed data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_HEAR:
	
	switch (data.sender) {
	case SENDER_SELF: 
	  TRACE(cout << "Comm: skipping own message" << endl);
	  break;
	case SENDER_REFEREE:
	  TRACE(cout << "Comm: skipping referee play mode" << endl);
	  // information contained was already used in info->adjust() above, 
	  // so also forget about it
	  break;
	case SENDER_DIR:    // message from player heard by player
	  TRACE(cout << "Comm: storing heard message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_PLAYER: // message from player heard by any coach
	  TRACE(cout << "Comm: storing player message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_COACH:  // message from train coach heard by player/olcoach
	  if (info->isTrainCoach()) {
	    TRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    TRACE(cout << "Comm: storing coach message relayed "
		  << "through the referee" << endl);
	    SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		     << data.message << endl);
	    audiomgr->storeMsg(AudioMsg(data.message));
	  }
	  break;
	case SENDER_ONLINECOACH:// msg from onlinecoach heard by player/coach
	  TRACE (cout << "Comm: recvd message from olcoach" << endl);
	  if (info->isOnlineCoach()) {
	    TRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    TRACE(cout << "Comm: storing online coach message" << endl);
	    switch (data.coachMsgType) {
	    case CMT_FREEFORM: {
	      SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		       << data.message << endl);
	      audiomgr->storeMsg(AudioMsg(data.message));
	      break;
	    }
	    case CMT_META:
	    case CMT_INFO:
	    case CMT_ADVICE:
	    case CMT_DEFINE:
	    default:;
	    } // switch on CMT
	  } // !isOnlineCoach
	  break;
	} // switch on sender
	break;
      } //switch on sensor_type
    } //parse ok
  } // while data

}


comm_error 
Comm::createSocket (const char *hostname, const int port) {
  // lookup the server's address, create a socket and bind it
  
  INITTRACE( cout << "Comm:     creating socket at " 
	     << hostname << ":" << port << endl);

  // Init remote and local addresses
  memset(&server_addr, 0, sizeof(sockaddr_in));
  server_addr.sin_family = AF_INET;

  struct sockaddr_in local_addr;
  memset(&local_addr, 0, sizeof(sockaddr_in));
  local_addr.sin_family = AF_INET ;

  INITTRACE( cout << "Comm:         looking network address for " 
	     << hostname << endl);
  // Look up the server's network address
  struct hostent *hp;
  if ((hp = gethostbyname(hostname)) != NULL)
    memcpy(&server_addr.sin_addr, hp->h_addr, sizeof(server_addr.sin_addr));
  else 
    if ((server_addr.sin_addr.s_addr = inet_addr(hostname)) == 
	(unsigned long)-1) return COMM_ERROR_SERVER_NOT_FOUND;
  
  server_addr.sin_port = htons(port);


  INITTRACE( cout << "Comm:         creating socket on remote port " 
	     << port << endl);
  // Create UDP socket
  if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
    return COMM_ERROR_SOCKET_INIT;

  // get SIGIO delivered to current process
  if (fcntl (sd, F_SETOWN, getpid())) return COMM_ERROR_SOCKET_INIT;
  
  // make socket non-blocking
  int val = fcntl(sd, F_GETFL, 0) ;
  //val |= O_ASYNC; // BSD-style option, does not work properly !
  val |= FASYNC; 
  // from CMU98: undocumented option, but the only one that works ;-)
  if (fcntl(sd, F_SETFL, val) ) return COMM_ERROR_SOCKET_INIT;
  
  INITTRACE( cout << "Comm:         binding local address to socket" << endl);
  // Bind local address
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  local_addr.sin_port = htons(0) ;
  if (bind(sd, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) {
    return COMM_ERROR_SOCKET_INIT;
  }
  INITTRACE( cout << "Comm:     socket sucessfully opened" << endl);
  
  return COMM_OK;

}

void
Comm::generateCheckinMessage(char* outmsg, 
			     client_type type, int reconnectNumber) {
  
  if (reconnectNumber <= 0) { // player first init or any coach
    switch (type) {
    case CLIENT_GOALIE:
      sprintf(outmsg, "(init %s (version 7.00) (goalie))", 
	      info->getTeam()); break;
    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(init %s (version 7.00))", info->getTeam()); break;
    case CLIENT_ONLINECOACH:
      sprintf(outmsg, "(init %s (version 7.00))", info->getTeam()); break;
    case CLIENT_TRAINCOACH:
      sprintf(outmsg, "(eye on)"); break;
    default: 
      break;
    }
  } else {
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(reconnect %s %d)", 
	      info->getTeam(), reconnectNumber); break;
    case CLIENT_ONLINECOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
    default:
      break;
    }
  } 

}

comm_error
Comm::parseCheckinResponse(char* inmsg, client_type type, 
			   int reconnectNumber, playing_side coachSide) {

  char   c;
  char   playmode [MAX_SERVER_MSG_SIZE];

  // needed to parse play mode returned by server
  Parser* dummy = new Parser("", 0, SIDE_UNKNOWN, false);

  int          time;
  bool         secondHalf;
  int          number = -1;
  play_mode    mode   = PLAY_MODE_UNKNOWN;
  playing_side side   = SIDE_UNKNOWN;
  
  if (reconnectNumber <= 0) { // parse the init response
    time = 0;
    secondHalf = false;
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(init %c %d %[^)])", &c, &number, playmode) != 3)
	return COMM_ERROR_PARSE;
      if (c == 'l') side = SIDE_LEFT;
      else if (c == 'r') side = SIDE_RIGHT;
      else return COMM_ERROR_PARSE;
      mode = dummy->parsePlayMode(playmode);
      break;
    case CLIENT_ONLINECOACH:
      if (sscanf(inmsg, "(init %c ok)" , &c) != 1) 
	return COMM_ERROR_PARSE;
      else {
	number = 0;
	if (c == 'l') side = SIDE_LEFT;
	else if (c == 'r') side = SIDE_RIGHT;
	else return COMM_ERROR_PARSE;
	mode = PLAY_MODE_BEFORE_KICK_OFF;
	sendEye(true);
      }
      break;
    case CLIENT_TRAINCOACH:
      if (strcmp(inmsg,"(ok eye on)")) return COMM_ERROR_PARSE;
      else {
	number = 0;
	side = coachSide;
	if (coachSide == SIDE_LEFT)
	  mode = PLAY_MODE_KICK_OFF_US;
	else 
	  mode = PLAY_MODE_KICK_OFF_THEM;
	sendEar(true);
      }
      break;
    default:
      break;
    }
  } else {  // parse the reconnect response
    time = 3000;
    secondHalf = true;
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(reconnect %c %[^)])", &c, playmode) != 2)
	return COMM_ERROR_PARSE;
      number = reconnectNumber;
      if (c == 'l') side = SIDE_LEFT;
      else if (c == 'r') side = SIDE_RIGHT;
      else return COMM_ERROR_PARSE;
      mode = dummy->parsePlayMode(playmode);
      break;

    case CLIENT_ONLINECOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;
      
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;
      
    default:
      break;
    }
  }
  
  info->setStates(type, number, time, mode, side, secondHalf);
  
  delete (dummy);
  return COMM_OK;

}

comm_error 
Comm::checkinAtServer (client_type type,
		       int reconnectNumber, playing_side coachSide) {
  
  INITTRACE( cout << "Comm:     ";
	     switch(type) {
	     case CLIENT_GOALIE: cout << "goalie "; break;
	     case CLIENT_FIELDPLAYER: cout << "fieldplayer "; break;
	     case CLIENT_ONLINECOACH: cout << "onlinecoach "; break;
	     case CLIENT_TRAINCOACH: cout << "trainingcoach "; break;
	     default: break;
	     }
	     if (reconnectNumber > 0) cout << "reconnecting" << endl;
	     else cout << "checking into server" << endl;);
  
  
  char outmsg[MAX_SERVER_MSG_SIZE];
  char inmsg[MAX_SERVER_MSG_SIZE];

  generateCheckinMessage (outmsg, type, reconnectNumber);
  
  // send message to server
  INITTRACE (cout << "Comm:     sending checkin message " 
	     << outmsg << " to server" << endl);
  if (sendToServer(outmsg) != COMM_OK) return COMM_ERROR_SERVER_INIT;
  
  INITTRACE( cout << "Comm:     waiting for server response" << endl);
  
  if (receiveFromServer(inmsg, COMM_MAX_DEAD_SERVER_CYCLES*SIMULATOR_STEP) 
      != COMM_OK)
    return COMM_ERROR_NO_DATA;
  INITTRACE (cout << "Comm:     received server response " 
	     << inmsg << endl);
  
  comm_error ce;
  ce = parseCheckinResponse (inmsg, type, reconnectNumber, coachSide);
  if (ce != COMM_OK) {
    CERR("Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): cannot handle server checkin response " << inmsg );
    return ce;
  }

  INITTRACE( cout << "Comm:     waiting for server_param" << endl);

  if (receiveFromServer(inmsg, COMM_MAX_DEAD_SERVER_CYCLES*SIMULATOR_STEP) 
      != COMM_OK)
    return COMM_ERROR_NO_DATA;
  INITTRACE (cout << "Comm:     received server_param " 
	     << inmsg << endl);
  
  ce = parseServerParam(inmsg);
  if (ce != COMM_OK) {
    CERR("Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): cannot parse server_param " << inmsg );
    return ce;
  }

  INITTRACE( cout << "Comm:     waiting for player_param" << endl);

  if (receiveFromServer(inmsg, COMM_MAX_DEAD_SERVER_CYCLES*SIMULATOR_STEP) 
      != COMM_OK)
    return COMM_ERROR_NO_DATA;
  INITTRACE (cout << "Comm:     received player_param " 
	     << inmsg << endl);
  
  ce = parsePlayerParam(inmsg);
  if (ce != COMM_OK) {
    CERR("Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): cannot parse player_param " << inmsg );
    return ce;
  }

  INITTRACE( cout << "Comm:     completed checkin at sserver" << endl);

  return COMM_OK;
  
}

comm_error 
Comm::initSigHandlers() {

  INITTRACE( cout << "Comm:     initializing signal handlers" << endl);

  struct sigaction saction;
  
  saction.sa_mask = alrmask; // SIGIO-exec blocks SIGALRM
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = Comm::ioHandler; //(void (*) (...)) Comm::ioHandler;
  if (sigaction (SIGIO, &saction, NULL)) return COMM_ERROR_SIGNAL_INIT;
  
  saction.sa_mask = iomask; // SIGALRM-exec blocks SIGIO
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = Comm::alarmHandler; //(void (*) (...)) Comm::alarmHandler;
  if (sigaction (SIGALRM, &saction, NULL)) return COMM_ERROR_SIGNAL_INIT;
  
  return COMM_OK;

}

comm_error 
Comm::setAlarmInterval(int msec) {
  
  INITTRACE(cout << "Comm:     setting alarm interval to " <<msec<< " msec" <<endl);

  long usec = msec*1000;
  
  itimerval itv;
  
  // time left in current interval
  itv.it_value.tv_sec     = usec / 1000000;  
  itv.it_value.tv_usec    = usec % 1000000;
  
  // stadard interval time
  itv.it_interval.tv_sec  = usec / 1000000;
  itv.it_interval.tv_usec = usec % 1000000;
  
  if (setitimer (ITIMER_REAL, &itv, NULL )) return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
  
}

void 
Comm::removeSigHandlers() {

  INITTRACE(cout << "Comm:     removing all signal handlers" << endl);

  struct sigaction saction;

  sigset_t emptymask;
  sigemptyset(&emptymask);

  saction.sa_mask = emptymask;
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = SIG_DFL;

  sigaction(SIGIO, &saction, NULL);
  sigaction(SIGALRM, &saction, NULL);
  
}

comm_error
Comm::blockSignals() {
  
  INITTRACE(cout << "Comm:     blocking all signals" << endl);

  if ( sigprocmask(SIG_BLOCK, &alrmask, NULL) ||
       sigprocmask(SIG_BLOCK, &iomask, NULL)  )
    return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
}

comm_error
Comm::unblockSignals() {

  INITTRACE(cout << "Comm:   unblocking all signals" << endl);

  if ( sigprocmask(SIG_UNBLOCK, &alrmask, NULL) ||
       sigprocmask(SIG_UNBLOCK, &iomask, NULL)  )
    return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
}


comm_error 
Comm::init(client_type type,
	   const char* hostname,
	   int port,
	   int reconnectNumber,
	   bool _printCommands,
	   playing_side coachSide) {
  
  INITTRACE(cout << "Comm: initializing and checking in" << endl);

  printCommands = _printCommands;
  
  comm_error e;
  
  e = blockSignals(); if(e) return e;
  
  e = createSocket (hostname, port); if (e) return e;
  
  e = checkinAtServer (type, reconnectNumber, coachSide); 
  if (e) return e;

  // set parameters known after receiving first info

  audiomgr->setStates(info->getNumber(), info->getSide()==SIDE_LEFT );
  
  parser->setStates(info->getTeam(), info->getNumber(), 
		    info->getSide(), info->isCoach());
  
  e = initSigHandlers(); if (e) return e;
  
  e = setAlarmInterval(SIMULATOR_STEP); if (e) return e;
  e = unblockSignals(); if(e) return e;

  serverOk = true;
  
  INITTRACE(cout << "Comm: init sucessfully completed" << endl);

  return COMM_OK;
}


bool 
Comm::getNextData(SensorData &data) {
  return store->getNext(data);
}

void
Comm::getLastAction(CycleAction &action) {
  info->getLastAction(action);
}

void 
Comm::handleSignals() {

#ifdef USE_PAUSE

  TRACE(cout << "Comm: pausing" << endl);
  pause();

#else

  sigset_t oldmask;
  sigset_t blockmask;
  
  sigemptyset(&blockmask);
  sigaddset(&blockmask, SIGIO);
  sigaddset(&blockmask, SIGALRM);
  
  TRACE(cout << "Comm: blocking signals" << endl);
  sigprocmask(SIG_BLOCK, &blockmask, &oldmask); // block all
  
  TRACE(cout << "Comm: suspending execution" << endl);
  sigsuspend(&oldmask);

#endif

  if (alarmFlag) { 
    alarmFlag = 0;
    finishAlarm();
  }
  
  if (ioFlag) {
    ioFlag = 0;
    finishIo();
  }

#ifdef USE_PAUSE
#else
  TRACE(cout << "Comm: unblocking signals" << endl);
  sigprocmask(SIG_UNBLOCK, &blockmask, NULL);
#endif

}

bool 
Comm::amIGoalie() {
  return info->isGoalie();
}

//////////////////////////////////////////////////////////////////////////


bool 
Comm::isServerAlive() {
  return serverOk;
}

bool
Comm::isGameOver() {
  return (info->getPlayMode() == PLAY_MODE_TIME_OVER);
}


int
Comm::getServerTime() {
  return info->getServerTime();
}

bool 
Comm::isTimeRunning() {
  return info->isTimeRunning();	   
}

bool 
Comm::isActionNeeded() {
  return info->actNow();
}

const char*  
Comm::getMyTeam() {
  return info->getTeam();
}

int 
Comm::getMyNumber() {
  return info->getNumber();
}

play_mode
Comm::getPlayMode() {
  return info->getPlayMode();
}

playing_side 
Comm::getMySide() {
  return info->getSide();
}

int
Comm::getOurScore() {
  return info->getOurScore();
}

int
Comm::getOtherScore() {
  return info->getOtherScore();
}

bool
Comm::inSecondHalf() {
  return info->inSecondHalf();
}

void
Comm::coachSay( void /*???*/) {

}

void
Comm::sendMCmd(const MCommand &cmd) {
  if (!Comm::theComm->FUNSSELmode) {
    CERR("Comm: Not in FUNSSELmode dropping cmd: \"" << cmd << "\"");
    return;
  }
  char buf[30];
  sprintf(buf, FILTER_STRING " %d ", cmd.duration);
  string msg(buf);
  msg += " " + (string)cmd + ")";
  if (Comm::theComm->sendToServer(msg.c_str()) != COMM_OK)
    CERR(Comm::theComm->getServerTime() 
	 << " Comm (" << Comm::theComm->info->getTeam() << " No." 
	 << Comm::theComm->info->getNumber()
	 << "): cannot send mcmd \"" << cmd << "\" to server." );
}

void
Comm::say(AudioMsg &msg) {

  //[C]
  char charmsg[MAX_SAY_MSG_SIZE];

  if (msg.isCoachMsg()) {
    strcpy(charmsg, "(freeform \""); 
    strcat(charmsg, audiomgr->SpeakTo(msg));
    strcat(charmsg, "\")");
  } else
    strcpy(charmsg, audiomgr->SpeakTo(msg));

  SAYTRACE(cout << "Comm: encoding via AudioMgr " << msg << endl);
  sendSay(charmsg);
  
}

bool
Comm::getNextMsg(AudioMsg &msg) {

  bool hadMsg = audiomgr->getNextMsg(msg);
  SAYTRACE(if (hadMsg) 
	   cout << "Comm: retrieved from AudioMgr " << msg << endl);
  return hadMsg;
  
}

void 
Comm::noAction() {

  TRACE(cout << "Comm: agent doing no action" << endl);
  info->setSentAction(NO_COMMAND, info->getServerTime(), 0, 0);

}

void 
Comm::sendKick    (double power, double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) && 
		 (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN somewhere ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(kick %f %f)", power, dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(KICK_COMMAND, info->getServerTime(), 
			power, DEG_TO_RAD( dir_deg ));
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendDash    (double power) { 

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(dash %f)", power);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(DASH_COMMAND, info->getServerTime(), power, 0);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendTurn    (double moment_rad) { 

  double moment_deg = RAD_TO_DEG( moment_rad );

  if (moment_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MINMOMENT;
  } else if (moment_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MAXMOMENT;
  } else if ( ! ( (moment_deg >= MINMOMENT - EPS) && 
		  (moment_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn %f)", moment_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(TURN_COMMAND, info->getServerTime(), 
			moment_deg * M_PI / 180.0, 0);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendCatch   (double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), dropping command completely" );
      return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(catch %f)", dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(CATCH_COMMAND, info->getServerTime(), 
			DEG_TO_RAD( dir_deg ), 0);
    info->incCatchCount();
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendMove    (double x, double y) { 

  if ( (!info->isGoalie() && info->isTimeRunning() ) ||
       (info->isGoalie() && info->isTimeRunning() && 
	(info->getPlayMode() != PLAY_MODE_FREE_KICK_US) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << ") game is running, agent not allowed to (move)" );
    return;
  } else {

    // no range checking when moving: there are no limits !
    
    char msg [MAX_SERVER_MSG_SIZE];
    sprintf(msg, "(move %f %f)", x, y);
    if (sendToServer(msg) == COMM_OK) {
      TRACE(cout << "Comm: agent sent " << msg << endl);
      info->setSentAction(MOVE_COMMAND, info->getServerTime(), x, y);
      info->incMoveCount();
    } else {
      CERR(getServerTime() 
	   << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): agent cannot send " << msg << " to server" );
    }
  }
  
}

void 
Comm::sendTurnNeck   (double angle_rad) { 

  double angle_deg = RAD_TO_DEG( angle_rad );

  if (angle_deg < MINNECKMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MINNECKMOMENT );
    angle_deg = MINNECKMOMENT;
  } else if (angle_deg > MAXNECKMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MAXNECKMOMENT );
    angle_deg = MAXNECKMOMENT;
  } else if (! ( (angle_deg >= MINNECKMOMENT - EPS) && 
		 (angle_deg <= MAXNECKMOMENT + EPS) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), dropping command completely" );
    return;
  }
  
  // to make sure there are no (see) infos left on the socket, just read 
  // and parse all outstanding IO into the SensorData buffer:
  TRACE(cout << "Comm: reading all IO just before turning neck" << endl);
  finishIo();
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn_neck %f)", angle_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE( cout << "Comm: agent sent " << msg << endl );
    
    info->incTurnNeckCount();

    /* no need to check for twisted neck, worldmodel can handle it anyway.
       
       info->twistNeck();
       
    */
    
    /* server handles turn_neck synchronously, no need to send sense_body !
       
       // if (see) info is received with a twisted neck an warning message 
       // will be printed. Only after a sense_body arrived with a confirmed 
       // head_angle will these warnings be removed. 
       // ( see AsyncInfo::adjust() and AsyncInfo::isNeckTwisted() )
       
       sendSenseBody();
    
    */
    
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendChangePlayer (int player, int type) {
  char s[50];
  sprintf(s, "(change_player_type %d %d)", player, type);
  if (sendToServer(s) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << s << endl);
  } else {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << s << " to server" );
  }
}

void 
Comm::sendChangeView (view_width width, view_quality quality) { 
  
  char *w = "";
  switch (width) {
  case VIEW_WIDTH_NARROW: w = "narrow"; break;
  case VIEW_WIDTH_NORMAL: w = "normal"; break;
  case VIEW_WIDTH_WIDE: w = "wide"; break;
  }

  char *q = "";
  switch (quality) {
  case VIEW_QUALITY_LOW: q = "low"; break;
  case VIEW_QUALITY_HIGH: q = "high"; break;
  }

  char msg[MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(change_view %s %s)", w, q);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    //    info->setSentAction(CHANGE_VIEW_COMMAND, info->getServerTime(), 
    //		width, quality);
    info->incChangeViewCount();
  } else {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendSenseBody  () {

  if (sendToServer("(sense_body)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (sense_body)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (sense_body) to server" );
  }
}

void 
Comm::sendSay(const char *message) { 

  unsigned int maxSize = MAX_SAY_MSG_SIZE;
  if (info->isCoach()) maxSize = SAY_COACH_MSG_SIZE + 11; 
  // + strlen("(freeform")+strlen(")") !!
  // [!] wie lang darf duerfen die messages hoechstens sein (+-10 Zeichen?)
  if (strlen(message) > maxSize) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): client tries to send too long say message (len= "
	 << strlen(message) << ")(msg= " << message 
	 << "), dropping it completely" );
    return;
  }
  
  char msg[maxSize];
  sprintf(msg, "(say %s)", message);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->incSayCount();
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}


void 
Comm::sendChangeMode(play_mode playMode) { 
  char* newMode;
  switch (playMode) {
  case PLAY_MODE_BEFORE_KICK_OFF: newMode = "before_kick_off"; break;
  case PLAY_MODE_CORNER_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "corner_kick_l";
    else newMode = "corner_kick_r";
    break;
  case PLAY_MODE_CORNER_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "corner_kick_r";
    else newMode = "corner_kick_l";
    break;
  case PLAY_MODE_FREE_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "free_kick_l";
    else newMode = "free_kick_r";
    break;
  case PLAY_MODE_FREE_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "free_kick_r";
    else newMode = "free_kick_l";
    break;
  case PLAY_MODE_GOAL_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_kick_l";
    else newMode = "goal_kick_r";
    break;
  case PLAY_MODE_GOAL_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_kick_r";
    else newMode = "goal_kick_l";
    break;
  case PLAY_MODE_KICK_IN_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_in_l";
    else newMode = "kick_in_r";
    break;
  case PLAY_MODE_KICK_IN_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_in_r";
    else newMode = "kick_in_l";
    break;
  case PLAY_MODE_KICK_OFF_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_off_l";
    else newMode = "kick_off_r";
    break;
  case PLAY_MODE_KICK_OFF_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_off_r";
    else newMode = "kick_off_l";
    break;
  case PLAY_MODE_GOAL_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_l";
    else newMode = "goal_r";
    break;
  case PLAY_MODE_GOAL_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_r";
    else newMode = "goal_l";
    break;
  case PLAY_MODE_OFFSIDE_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "offside_l";
    else newMode = "offside_r";
    break;
  case PLAY_MODE_OFFSIDE_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "offside_r";
    else newMode = "offside_l";
    break;
  case PLAY_MODE_PLAY_ON : newMode = "play_on"; break;
  case PLAY_MODE_TIME_OVER : newMode = "time_over"; break;
  default: 
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): trying to set unknown play mode: No." 
	 << (int)playMode << ", droping command" ); 
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(change_mode %s)", newMode);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y) { 
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f)", object, x, y);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y, 
		      double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f)", object, x, y, dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y, 
		      double dir_rad, double xvel, double yvel) { 
  
  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f %f %f)", object, x, y, dir_deg, xvel, yvel);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendEye        (bool eyeOn) {
  
  char* msg;
  if (eyeOn) msg = "(eye on)";
  else msg = "(eye off)";
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendEar        (bool earOn) { 
  char* msg;
  if (earOn) msg = "(ear on)";
  else msg = "(ear off)";
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendLook       () { 
  if (sendToServer("(look)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (look)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (look) to server" );
  }
}  

void 
Comm::sendCheckBall  () { 
  if (sendToServer("(check_ball)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (check_ball)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (check_ball) to server" );
  }
}

void 
Comm::sendRecover  () { 
  if (sendToServer("(recover)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (recover)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (recover) to server" );
  }
}

comm_error
Comm::parseServerParam(char* s) {
  TRACE(cout << "Comm:    parsing server_param" << endl);
  PARAMTRACE("Serverparams", "...");
  int len(0), l(0);
  double dummy;
  if (sscanf(s, "(server_param %lf %n", &GOAL_WIDTH, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " GOAL_WIDTH ", GOAL_WIDTH );
  if (sscanf(s + len, "%lf %n", &INERTIA_MOMENT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " INERTIA_MOMENT ", INERTIA_MOMENT );
  if (sscanf(s + len, "%lf %n", &PLAYER_SIZE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_SIZE ", PLAYER_SIZE );
  if (sscanf(s + len, "%lf %n", &PLAYER_DECAY, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_DECAY ", PLAYER_DECAY );
  if (sscanf(s + len, "%lf %n", &PLAYER_RAND, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_RAND ", PLAYER_RAND );
  if (sscanf(s + len, "%lf %n", &PLAYER_WEIGHT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_WEIGHT ", PLAYER_WEIGHT );
  if (sscanf(s + len, "%lf %n", &PLAYER_SPEED_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_SPEED_MAX ", PLAYER_SPEED_MAX );
  if (sscanf(s + len, "%lf %n", &PLAYER_ACCEL_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE( " PLAYER_ACCEL_MAX ", PLAYER_ACCEL_MAX );
  if (sscanf(s + len, "%lf %n", &STAMINA_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" STAMINA_MAX ",  STAMINA_MAX );
  if (sscanf(s + len, "%lf %n", &STAMINA_INC_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" STAMINA_INC_MAX ",  STAMINA_INC_MAX );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.recover_init*/, &l)
      != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &RECOVER_DEC_THR, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" RECOVER_DEC_THR ",  RECOVER_DEC_THR );
  if (sscanf(s + len, "%lf %n", &RECOVER_MIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" RECOVER_MIN ",  RECOVER_MIN );
  if (sscanf(s + len, "%lf %n", &RECOVER_DEC, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" RECOVER_DEC ",  RECOVER_DEC );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.effort_init*/, &l)
      != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &EFFORT_DEC_THR, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" EFFORT_DEC_THR ",  EFFORT_DEC_THR );
  if (sscanf(s + len, "%lf %n", &EFFORT_MIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" EFFORT_MIN ",  EFFORT_MIN );
  if (sscanf(s + len, "%lf %n", &EFFORT_DEC, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" EFFORT_DEC ",  EFFORT_DEC );
  if (sscanf(s + len, "%lf %n", &EFFORT_INC_THR, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" EFFORT_INC_THR ",  EFFORT_INC_THR );
  if (sscanf(s + len, "%lf %n", &EFFORT_INC, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" EFFORT_INC ",  EFFORT_INC );
  if (sscanf(s + len, "%lf %n", &KICK_RAND, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" KICK_RAND ",  KICK_RAND );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.team_actuator_noise*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.prand_factor_l*/
	     , &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.prand_factor_r*/, 
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.kick_rand_factor_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.kick_rand_factor_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &BALL_SIZE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BALL_SIZE ",  BALL_SIZE );
  if (sscanf(s + len, "%lf %n", &BALL_DECAY, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BALL_DECAY ",  BALL_DECAY );
  if (sscanf(s + len, "%lf %n", &BALL_RAND, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BALL_RAND ",  BALL_RAND );
  if (sscanf(s + len, "%lf %n", &BALL_WEIGHT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BALL_WEIGHT ",  BALL_WEIGHT );
  if (sscanf(s + len, "%lf %n", &BALL_SPEED_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BALL_SPEED_MAX ",  BALL_SPEED_MAX );
  if (sscanf(s + len, "%lf %n", &BACCEL_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" BACCEL_MAX ",  BACCEL_MAX );
  if (sscanf(s + len, "%lf %n", &DASH_POWER_RATE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" DASH_POWER_RATE ",  DASH_POWER_RATE );
  if (sscanf(s + len, "%lf %n", &KICK_POWER_RATE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" KICK_POWER_RATE ",  KICK_POWER_RATE );
  if (sscanf(s + len, "%lf %n", &KICKABLE_MARGIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" KICKABLE_MARGIN ",  KICKABLE_MARGIN );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.ctlradius*/, 
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.ctlradius_width*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &MAXPOWER, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MAXPOWER ",  MAXPOWER );
  if (sscanf(s + len, "%lf %n", &MINPOWER, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MINPOWER ",  MINPOWER );
  if (sscanf(s + len, "%lf %n", &MAXMOMENT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MAXMOMENT ",  MAXMOMENT );
  if (sscanf(s + len, "%lf %n", &MINMOMENT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MINMOMENT ",  MINMOMENT );
  if (sscanf(s + len, "%lf %n", &MAXNECKMOMENT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MAXNECKMOMENT ",  MAXNECKMOMENT );
  if (sscanf(s + len, "%lf %n", &MINNECKMOMENT, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MINNECKMOMENT ",  MINNECKMOMENT );
  if (sscanf(s + len, "%lf %n", &MAXNECKANG, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MAXNECKANG ",  MAXNECKANG );
  if (sscanf(s + len, "%lf %n", &MINNECKANG, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" MINNECKANG ",  MINNECKANG );
  if (sscanf(s + len, "%lf %n", &VISIBLE_ANGLE, &l) != 1)
    return COMM_ERROR_PARSE;
  else 
    VISIBLE_ANGLE = DEG_TO_RAD(VISIBLE_ANGLE);
  len += l;
  PARAMTRACE(" VISIBLE_ANGLE ",  VISIBLE_ANGLE );
  if (sscanf(s + len, "%lf %n", &VISIBLE_DISTANCE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" VISIBLE_DISTANCE ",  VISIBLE_DISTANCE );
  if (sscanf(s + len, "%lf %n", &WIND_DIR, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" WIND_DIR ",  WIND_DIR );
  if (sscanf(s + len, "%lf %n", &WIND_FORCE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" WIND_FORCE ",  WIND_FORCE );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.winang*/, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &WIND_RAND, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" WIND_RAND ",  WIND_RAND );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.kickable_area*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &CATCHABLE_AREA_L, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CATCHABLE_AREA_L ",  CATCHABLE_AREA_L );
  if (sscanf(s + len, "%lf %n", &CATCHABLE_AREA_W, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CATCHABLE_AREA_W ",  CATCHABLE_AREA_W );
  if (sscanf(s + len, "%lf %n", &CATCH_PROBABILITY, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CATCH_PROBABILITY ",  CATCH_PROBABILITY );
  if (sscanf(s + len, "%d %n", &GOALIE_MAX_MOVES, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" GOALIE_MAX_MOVES ",  GOALIE_MAX_MOVES );
  if (sscanf(s + len, "%lf %n", &CKICK_MARGIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CKICK_MARGIN ",  CKICK_MARGIN );
  if (sscanf(s + len, "%lf %n", &OFFSIDE_ACTIVE_AREA, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" OFFSIDE_ACTIVE_AREA ",  OFFSIDE_ACTIVE_AREA );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.win_no*/, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &WIND_RAND, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" WIND_RAND ",  WIND_RAND );
  if (sscanf(s + len, "%d %n", &SAY_COACH_CNT_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SAY_COACH_CNT_MAX ",  SAY_COACH_CNT_MAX );
  if (sscanf(s + len, "%d %n", &SAY_COACH_MSG_SIZE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SAY_COACH_MSG_SIZE ",  SAY_COACH_MSG_SIZE );
  if (sscanf(s + len, "%d %n", &CLANG_WIN_SIZE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_WIN_SIZE ",  CLANG_WIN_SIZE );
  if (sscanf(s + len, "%d %n", &CLANG_DEFINE_WIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_DEFINE_WIN ",  CLANG_DEFINE_WIN );
  if (sscanf(s + len, "%d %n", &CLANG_META_WIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_META_WIN ",  CLANG_META_WIN );
  if (sscanf(s + len, "%d %n", &CLANG_ADVICE_WIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_ADVICE_WIN ",  CLANG_ADVICE_WIN );
  if (sscanf(s + len, "%d %n", &CLANG_INFO_WIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_INFO_WIN ",  CLANG_INFO_WIN );
  if (sscanf(s + len, "%d %n", &CLANG_MESS_DELAY, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_MESS_DELAY ",  CLANG_MESS_DELAY );
  if (sscanf(s + len, "%d %n", &CLANG_MESS_PER_CYCLE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CLANG_MESS_PER_CYCLE ",  CLANG_MESS_PER_CYCLE );
  if (sscanf(s + len, "%d %n", &HALF_TIME, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" HALF_TIME ",  HALF_TIME );
  if (sscanf(s + len, "%d %n", &SIMULATOR_STEP, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SIMULATOR_STEP ",  SIMULATOR_STEP );
  if (sscanf(s + len, "%d %n", &SEND_STEP, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SEND_STEP ",  SEND_STEP );
  if (sscanf(s + len, "%d %n", &RECV_STEP, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" RECV_STEP ",  RECV_STEP );
  if (sscanf(s + len, "%d %n", &SENSE_BODY_STEP, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SENSE_BODY_STEP ",  SENSE_BODY_STEP );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.lcm_st*/, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%d %n", &SAY_MSG_SIZE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SAY_MSG_SIZE ",  SAY_MSG_SIZE );
  if (sscanf(s + len, "%lf %n", &HEAR_MAX, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" HEAR_MAX ",  HEAR_MAX );
  if (sscanf(s + len, "%lf %n", &HEAR_INC, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" HEAR_INC ",  HEAR_INC );
  if (sscanf(s + len, "%lf %n", &HEAR_DECAY, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" HEAR_DECAY ",  HEAR_DECAY );
  if (sscanf(s + len, "%d %n", &CATCH_BAN_CYCLE, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" CATCH_BAN_CYCLE ",  CATCH_BAN_CYCLE );
  if (sscanf(s + len, "%d %n", &SLOW_DOWN_FACTOR, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SLOW_DOWN_FACTOR ",  SLOW_DOWN_FACTOR ); 
  short b(0);
  if (sscanf(s + len, "%hd %n", &b, &l) != 1)
    return COMM_ERROR_PARSE;
  else
    USE_OFFSIDE = !!b;
  len += l;
  PARAMTRACE(" USE_OFFSIDE ",  USE_OFFSIDE );
  if (sscanf(s + len, "%hd %n", &b, &l) != 1)
    return COMM_ERROR_PARSE;
  else
    FORBID_KICK_OFF_OFFSIDE = !!b;
  len += l;
  PARAMTRACE(" FORBID_KICK_OFF_OFFSIDE ",  FORBID_KICK_OFF_OFFSIDE );
  if (sscanf(s + len, "%lf %n", &OFFSIDE_KICK_MARGIN, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" OFFSIDE_KICK_MARGIN ",  OFFSIDE_KICK_MARGIN );
  if (sscanf(s + len, "%lf %n", &AUDIO_CUT_DIST, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" AUDIO_CUT_DIST ",  AUDIO_CUT_DIST );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dist_qstep*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.land_qstep*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dir_qstep*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dist_qstep_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dist_qstep_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.land_qstep_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.land_qstep_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dir_qstep_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.dir_qstep_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.coach_mode*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.c_w_r_mode*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.old_hear*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%d %n", &SEND_VI_STEP, &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" SEND_VI_STEP ",  SEND_VI_STEP );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.start_goal_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.start_goal_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.fullstate_l*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%lf %n", &dummy/*&param.fullstate_r*/,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  len += l;
  PARAMTRACE(" dummy ",  dummy );
  if (sscanf(s + len, "%d %n", &DROP_BALL_TIME,
	     &l) != 1)
    return COMM_ERROR_PARSE;
  PARAMTRACE(" DROP_BALL_TIME ",  DROP_BALL_TIME );
  return COMM_OK;
}

comm_error
Comm::parsePlayerParam(char* s){
  TRACE(cout << "Comm:    parsing player_param" << endl);
  if (sscanf(s, "(player_param %d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf)",
	     &playerParams.player_types,
	     &playerParams.subs_max,
	     &playerParams.pt_max, 
	     &playerParams.player_speed_max_delta_min,
	     &playerParams.player_speed_max_delta_max, 
	     &playerParams.stamina_inc_max_delta_factor,
	     &playerParams.player_decay_delta_min, 
	     &playerParams.player_decay_delta_max,
	     &playerParams.inertia_moment_delta_factor, 
	     &playerParams.dash_power_rate_delta_min,
	     &playerParams.dash_power_rate_delta_max, 
	     &playerParams.player_size_delta_factor,
	     &playerParams.kickable_margin_delta_min,
	     &playerParams.kickable_margin_delta_max,
	     &playerParams.kick_rand_delta_factor, 
	     &playerParams.extra_stamina_delta_min, 
	     &playerParams.extra_stamina_delta_max, 
	     &playerParams.effort_max_delta_factor, 
	     &playerParams.effort_min_delta_factor) == 19) 
    return COMM_OK;

  return COMM_ERROR_PARSE;
}

