//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: datastore.C,v 1.5 2001/03/10 17:26:50 satan Exp $
// 
//  datastore.C - storing of SensorData
// 
//////////////////////////////////////////////////////////////////////


#include "datastore.h"

#undef TRACE
#define TRACE(cmd)


//////////////////////////////////////////////////////////////////////


SensorDataStore::SensorDataStore() {

  TRACE (cout << "DataStore: entering constructor" << endl);
  ensure (DATA_STORE_SIZE > 0);
  readIndex = 0;
  writeIndex = 0;
  TRACE (cout << "DataStore: leaving constructor" << endl);

}

SensorDataStore::~SensorDataStore() {
  TRACE (cout << "DataStore: entering destructor" << endl);
  TRACE (cout << "DataStore: leaving destructor" << endl);
}


//////////////////////////////////////////////////////////////////////////


void
SensorDataStore::insert(const SensorData &data) {

  TRACE(cout << "DataStore: trying to insert SensorData.. " << endl);
  
  int insertIndex = writeIndex;
  while ( (insertIndex > readIndex) && 
	  (store[insertIndex-1].time >= data.time) ) {
    TRACE(cout << "DataStore: moving old data from Index " << insertIndex-1 
	  << " to Index " << insertIndex << endl);
    store[insertIndex] = store[insertIndex-1];
    insertIndex--;
  }
  TRACE(cout << "DataStore: inserting at Index " << insertIndex << endl);
  store[insertIndex] = data;

  writeIndex++; // make room for next write access
  writeIndex = writeIndex % DATA_STORE_SIZE; // wrap around
  
  if (writeIndex == readIndex) { 
    // buffer full, next writeIndex caught up to readIndex from behind !
    readIndex++;
    readIndex = readIndex % DATA_STORE_SIZE;
    
    // In this case, the woldmodel was not called to adjust itself for 
    // too long a time, and the data coming in from the server was not
    // processed. If this happens, typically the call to Worldmodel::adjust() 
    // was "forgotten" in the agent code, or the machine running the agent 
    // is DEAD SLOW. Serious trouble.
    cerr << "DataStore: more than " << DATA_STORE_SIZE
	 << " unprocessed SensorDatas, dumping oldest one." << endl;

  }

}
  

void
SensorDataStore::put(const SensorData &data) {

  TRACE(cout << "DataStore: trying to store SensorData.. ");
  
  TRACE(cout << "writing to Index " << writeIndex << endl);
  
  store[writeIndex] = data;
  writeIndex++; // make room for next write access
  writeIndex = writeIndex % DATA_STORE_SIZE; // wrap around
  
  if (writeIndex == readIndex) { 
    // buffer full, next writeIndex caught up to readIndex from behind !
    readIndex++;
    readIndex = readIndex % DATA_STORE_SIZE;
    
    // In this case, the woldmodel was not called to adjust itself for 
    // too long a time, and the data coming in from the server was not
    // processed. If this happens, typically the call to Worldmodel::adjust() 
    // was "forgotten" in the agent code, or the machine running the agent 
    // is DEAD SLOW. Serious trouble.
    cerr << "DataStore: more than " << DATA_STORE_SIZE
	 << " unprocessed SensorDatas, dumping oldest one." << endl;

  }

}


bool 
SensorDataStore::getNext(SensorData &data) {

  TRACE(cout << "DataStore: trying to unqueue SensorData.. ");
  
  bool hadData;

  if (readIndex == writeIndex) {

    hadData = false;  
    TRACE(cout << "queue empty" << endl);
    
  } else {
    
    hadData = true;  
    TRACE (cout << "reading from Index " << readIndex << endl);

    data = store[readIndex];
    readIndex++;
    readIndex = readIndex % DATA_STORE_SIZE; //wrap around
    
  } 

  return hadData;

}
