//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-/////////////////////////////////////////////////////////
//
// $Id: onlinecoach.h,v 1.13 2001/06/04 14:26:02 satan Exp $
//
// onlinecoach.h - interface for online coach client
//
///////////////////////////////////////////////////////////////////////


#ifndef _ONLINECOACH_H_
#define _ONLINECOACH_H_

#include <unistd.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>


#include "constants.h"
#include "comm.h"
#include "worldmodel.h"
#include "audiomgr.h"

// filename of snapshot log files
#define SNAP_LOG_FILENAME "snaps_"
#define SNAP_LOG_EXTENSION ".log"

class Onlinecoach {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in a onlinecoach client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.
  //
  // Parameter description:
  // - numSnapshots
  // Specifies the memory usage of the Worldmodel. Larger numbers will keep
  // more snapshots of the past accessible, but consume quite some memory.

  Onlinecoach (const char*  team_name,            // name of own team
	       const char*  host             = DEFAULT_HOST,
	       int          port             = OLCOACH_PORT,   
	       bool         logging_snaps    = false,
	       bool         FUNSSELmode      = false,
	       int          numSnapshots     = DEFAULT_SNAPSHOT_DEPTH);


  // Starts execution of the Onlinecoach client. 
  // After calling this method, the agent will repeatedly call act(snap)
  // every time it needs to send an action while the server is running.
  //
  // You cannot override this method.
  void coach();

  
  // Called once every time the online coach can say something to the team,
  // see also analyze() below.
  //
  // Override this method virtual in your derived classes to implement 
  // onlinecoach behaviour.
  virtual void act(Snapshot* snap) = 0;

  // Called once every timestep during the game. If the game timer is 
  // stopped, it is repeatedly called once every SIMULATOR_STEP ms.
  // It is always called _before_ act() for the relevant timestep.
  //
  // Override this method virtual in your derived classes to implement 
  // onlinecoach behaviour.
  virtual void analyze(Snapshot* snap) = 0;

  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Onlinecoach();  


  
  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();}
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}

  // returns the timestep the server is in, our current time.
  // (There is only ONE current time. This is it.)
  int          getTime()      {return comm->getServerTime();} 

  // if needed, returns a snapshot for a given time from the worldmodel
  Snapshot*    getSnapshot(int time) {return world->getSnapshot(time);}
  
  // sends an audiomessage to other clients immediately
  void         say(AudioMsg &msg)         {comm->say(msg);}

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  // send a change player type msg to sserver
  void sendChangePlayer(int player, int type) 
  { comm->sendChangePlayer(player, type); };

private:

  Comm *comm;

  Worldmodel *world;

  bool writeSnaps;
  ofstream out;

};


#endif
