//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// $Id: ballhandlingEvaluate.C,v 1.95 2001/08/06 02:56:13 mrb Exp $

// cwz-build-comman: ssh chakotay "cd /nhomes/janeway/axel/RoboCup_2000; make"
// cwz-build-command: (cd ../.. ; make)

/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////


Grade 
BallHandlingModule::evaluate(AdvSkills *compat_skills, 
			     const Snapshot *compat_snap, 
			     TacticalSnapshot *compat_tsnap) {
  // main evaluate method, called by master control to evaluate
  // current situation by module and return a grade of responsibility
  // and desired activity. Naturally plans some of the future actions
  // already in its computations.
  
  acted = false;
  findBallTurnedBody = false;

#ifdef BALLHTRACE
  strstream tmp;
  tmp << "tactical -> recvd state = " << tsnap->ourAction.state
      << ", performer = " << tsnap->ourAction.performer
      << ", receiver = " << tsnap->ourAction.receiver
      << ", 1. resp = " << tsnap->firstResponsiblePlayer
      << ", 2. resp = " << tsnap->secondResponsiblePlayer
      << ", 3. resp = " << tsnap->thirdResponsiblePlayer
      << (char)0;
  BHTRC(MLog(tmp.str()));

  strstream tmp2;
  tmp2 << "tactical -> reachT = " << tsnap->ourTeamReachBallTime(my.number)
       << " (" << tsnap->ourTeamReachBallTime(tsnap->firstResponsiblePlayer)
       << ", " << tsnap->ourTeamReachBallTime(tsnap->secondResponsiblePlayer)
       << ", " << tsnap->ourTeamReachBallTime(tsnap->thirdResponsiblePlayer)
       << ")"
       << (char)0;
  BHTRC(MLog(tmp2.str()));
#endif

  // check wether to turn the head to the ball
  // done here since we ALWAYS need the turnHeadGrade
  // and it has to be done before turnHeadEvaluate to allow
  // physical turns to be done.
  Grade lookG = evaluateFindBall();

  // check for standard situation -> perform FIND_BALL if needed
  if ( playMode != PLAY_MODE_PLAY_ON ) {
    // special case: GOAL_KICK_US : Goalie should be able to catch
    // in case anything unexpected happens
    if( playMode != PLAY_MODE_GOAL_KICK_US || snap->my.number != snap->ourGoalieNo )
      {
	action = FIND_BALL;
	grade = lookG;
	GRADETRC("eval -> only FIND_BALL possible, planning to let standard sit have control", grade);
	return grade;
      }
  }

  if (lookG.is_PERFECT()) {
    // this is only possible if ball data is corrupt
    // in which case we cannot do anything else
    action = FIND_BALL;
    return lookG;
  }

  grade = (goalie) ? evaluateGoalie() :
    evaluatePlayer();
  if (grade.is_worse_than(lookG)) {
    action = FIND_BALL;
    GRADETRC("eval -> ball too old, planning to look for ball",lookG);
    return lookG;
  }
  // looking is not really needed, so do normal action
  return grade;
}

Grade 
BallHandlingModule::turnHeadEvaluate(AdvSkills *compat_skills, 
				     AdvSkills::TurnHeadCommand &thc)
{
  Grade g;
  g.set_IMPOSSIBLE();
  // IMPOSSIBLE here means: use turnHeadGrade
  switch (action) {
  case WAIT:
    break;
  case FIND_BALL:
    if (acted) {
      // turning to search ball doesnt make sense if anyone else turns the Head
      g.set_PERFECT();
      // if findBall performed a turn command, follow that with the head
      if( findBallTurnedBody )
	// this considers if the player is too fast to do the full turn
	thc = skills->turnHead( snap, snap->my.bodyDir + turnAngle / snap->my.inertiaMoment() );
      else
	thc = skills->turnHead( snap, snap->my.bodyDir );
    }
    break;
  case GOALIE_INTERCEPT:
  case PLAYER_INTERCEPT:
  case PLAYER_INTERCEPT_ALT:
  case PLAYER_PLACE:
  case GOALIE_PLACE:
    BHTRC(MLog("th -> place/icpt"));
    if (acted) {
      // in principle, look onto ball
      g = grade;
      thc = skills->turnHead(snap, snap->ball);
      thc.vwidth = VIEW_WIDTH_NARROW;
      if (distToBall < 3*PLAYER_SPEED_MAX) {
	// if I am colliding, look onto ball
	// more or less a work around
	if ((tsnap->firstDangerousOpponent != UNKNOWN) &&
	    (my.distTo(snap->otherTeam(tsnap->firstDangerousOpponent))
	     < 3*PLAYER_SIZE)) {
	  g.set_PERFECT();
	  thc = skills->turnHead(snap, ball);
	}
	else {
	  // try to get a picture of the dribbleDir, if not current
	  const DribbleAreaData *adt = &areas[dribbleArea()];
	  if ((snap->time - snap->visualTime > 1) ||
	      (fabs(adt->prefHeadDir - snap->my.faceDir) > PI_4)) {
	    BHTRC(MLog("th -> look in dribbleDir"));
	    thc = skills->turnHead(snap, adt->prefHeadDir);
	    // highest grade possible
	    g = adt->dirs[0].grade;
	    g.downgrade();
	  }
	}
      }
    }
    break;
  case PLAYER_DRIBBLE:
  case PLAYER_FAST_DRIBBLE:
  case GOALIE_DRIBBLE:
    BHTRC(MLog("th -> dribble"));
    // turn head into dribble direction
    // turnHeadToBall doesn't make much sense
    BHTRC(LINE(my.pos, my.pos + radAng2Vector2d(1.5, dribbleDir)));
    BHTRC(TEXT("headdir", my.pos + radAng2Vector2d(1.5, dribbleDir)));
    thc = skills->turnHead(snap, dribbleDir);
    if (distToBall < 1.5*skills->kickRange) {
      thc.vwidth = VIEW_WIDTH_NARROW;
    }
    // dribbling gets better if we look into the dribble
    // direction. So try looking there even if we do not
    // actually dribble, but with less grade
    g = grade;
    if (!acted)
      g.downgrade();
    else {
      // look into dribble dir every second cycle at least
      if (fabs(snap->my.faceDir - dribbleDir) > PI_4)
	g.set_PERFECT();
    }
    break;
  case GOALIE_CATCH:
    break;
  }

  if (turnHeadGrade.is_better_than(g)) {
    // looking to ball is the optimum
    thc = skills->turnHead(snap, snap->ball);
    thc.vwidth = reqWidth;
    return turnHeadGrade;
  }
  return g;
}


Grade
BallHandlingModule::evaluateGoalie()
{
  // evaluate goalie's possible actions
  Grade g;
  
  // check if the situation is dangerous
  bool danger = 
    ( snap->ball.pos.getX() < -FIELD_LENGTH / 4.0 ) &&   // ball near our goal
    ( tsnap->ourAction.state != TA_PASS ||               // do not run after kickoffs!
      tsnap->ourAction.performer != snap->my.number );

  // check point where the ball could be stopped by an opponent
  WorldBall simBall = snap->ball;
  if( tsnap->firstDangerousOpponent != UNKNOWN )
    for( int i = 0; i <= tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent ); i++ )
      simBall.simulate();
  else
    simBall.pos += simBall.v * ( 1.0 / ( 1.0 - BALL_DECAY ) );

  // if this point is far from the goal (e.g. after goalkick) there is no direct danger
  if( simBall.pos.getX() > -FIELD_LENGTH / 4.0 )
    danger = false;
  
  // only get active in dangerous situations
  if( danger )
    {
      // try to catch ball, if possible -> PERFECT
      catchParams = skills->goalieCatch( snap );
      if( catchParams.state == AdvSkills::COK )
	{
	  // check if near border of catchable area
	  if( distToBall > 0.9 * CATCHABLE_AREA_L )
	    {
	      WorldBall wb;
	      // ok, catch is unsafe, check if we can wait
	      AdvSkills::TurtleCommand turtleParams =
		skills->controlBall( snap, 2 * MAXPOWER, 0.8 * CATCHABLE_AREA_L, wb );
	      // one dash helps and no opponent is too near
	      if( ( turtleParams.time == 1 ) &&
		  ( ( tsnap->firstDangerousOpponent == UNKNOWN ) ||
		    ( tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent ) > 1 ) ) )
		{
		  // check if ball will roll into the goal in the next timestep
		  WorldBall nextBall = ball;
		  nextBall.simulate();
		  // check for ball safely in penalty area next step
		  if( ( simBall.pos.getX() > -0.5 * FIELD_LENGTH + BALL_SIZE ) &&
		      ( simBall.pos.getX() < -0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - BALL_SIZE ) &&
		      ( fabs( simBall.pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - BALL_SIZE ) )
		    {
		      // another dash makes catching safer, so intercept
		      action = GOALIE_INTERCEPT;
		      g.set_PERFECT();
		      GRADETRC("eval -> catch unsafe, going near ball", g);
		      return g;
		    }
		}
	      // here for some reason a security dash was impossible, so better try unsafe catch
	    }
	  action = GOALIE_CATCH;
	  g.set_PERFECT();
	  GRADETRC("eval -> planning to catch ball", g);
	  return g;
	} // catching was ok

      // try to dribble ball back into penalty area or to place it
      if( skills->kickSafe( snap ) )
	{ 
	  // ball is kickable
	  g = goalieEvaluateDribble();

	  GRADETRC("eval -> cannot catch ball, planning to dribble/place", g);
	  if( g.is_BAD() )
	    BHTRC(MLog("dribbling/placing is BAD -> passmodule should do something!"));

	  return g;
	}

      // I am most responsible but the ball is not in kick/catch range
      if ( tsnap->firstResponsiblePlayer == my.number )
	{
	  // there is no dangerous opponent around
	  if( ( tsnap->firstDangerousOpponent == UNKNOWN ) ||
	      ( tsnap->ourTeamReachBallTime( my.number ) +
		// be more careful if ball is not in penalty area
		( snap->ball.inOurPenaltyArea() ? 0 : 8 ) <
		tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent ) ) )
	    {
	      // go for the ball
	      action = GOALIE_INTERCEPT;
	      g.set_PERFECT();
	      GRADETRC("eval -> closest to ball, planning to intercept it", g);
	      return g;
	    }
	}

      // if there is a goalkick, intercept
      Vector2d ballEndPos = snap->ball.pos + snap->ball.v * ( 1.0 / ( 1.0 - BALL_DECAY ) );
      if( ballEndPos.getX() < -0.5 * FIELD_LENGTH )
	{
	  // check if ball crosses the line between the goal posts
	  double factor = - ( 0.5 * FIELD_LENGTH + snap->ball.pos.getX() ) /
	    ( ballEndPos.getX() - snap->ball.pos.getX() );
	  if( fabs( snap->ball.pos.getY() + factor *
		    ( ballEndPos.getY() - snap->ball.pos.getY() ) ) <= 0.5 * GOAL_WIDTH )
	    {
	      GRADETRC("eval-> ball will cross goal line, planning to intercept it", g);
	      action = GOALIE_INTERCEPT;
	      g.set_PERFECT();
	      return g;
	    }
	}

      // in all other cases positioning should do the job!
      action = GOALIE_INTERCEPT;
      g.set_BAD();
      GRADETRC("eval -> planning to intercept, but positioning should do the job!", g);

      return g;
    }

  // no danger - positioning should do the job
  action = WAIT;
  g.set_BAD();
  GRADETRC("eval -> no danger, positioning should do the job!", g);

  return g;
}


Grade
BallHandlingModule::goalieEvaluateDribble() {
  // possibly place ball away from opponent, or dribble back into penalty area 
  // to have ball safe

  Grade g;
  
  if (ball.inOurPenaltyArea()) {
    action = GOALIE_PLACE;
    // placing is perfect if we can be sure that the opponent can not reach us
    // before we can catch again
    if ( (tsnap->firstDangerousOpponent == UNKNOWN) ||
	 (tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) >
	  max( 0, CATCH_BAN_CYCLE - tsnap->time + snap->lastCatchTime ) + 3)){
      placeDir = 0;
      BHTRC(MLog("dribble -> placing ball away from goal "
		      "(opponent is far away, placing should be safe)"));
      g.set_PERFECT();
    } else {
      WorldPlayer first = snap->otherTeam(tsnap->firstDangerousOpponent);
      
      if ( (tsnap->secondDangerousOpponent != UNKNOWN) &&
	   (tsnap->otherTeamReachBallTime(tsnap->secondDangerousOpponent) < 5)) {
	WorldPlayer second = snap->otherTeam(tsnap->secondDangerousOpponent);
	placeDir = bisect ( (my.pos - first.pos).getAngle(),
			    (my.pos - second.pos).getAngle() );
	BHTRC(MLog("dribble -> placing opposite of both closest opponents (placing is dangerous!)"));
      }
      else {
	placeDir = (my.pos - first.pos).getAngle();
	BHTRC(MLog("dribble -> placing opposite of closest opponent (placing is dangerous!)"));
      }
	
      // placing is always bad if opponent is close!
      g.set_BAD();
    }
  } else {
    // dribble into penalty area
    action = GOALIE_DRIBBLE;
    Vector2d point;
    if (ball.pos.getX() > - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - 1) {
      point.setX(- 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - 1);
      if (ball.pos.getY() < - 0.5 * PENALTY_AREA_WIDTH + 1) {
	point.setY(- 0.5 * PENALTY_AREA_WIDTH + 1);
      } else if (ball.pos.getY() > 0.5 * PENALTY_AREA_WIDTH - 1) {
	point.setY(0.5 * PENALTY_AREA_WIDTH - 1); 
      } else {
	point.setY(my.pos.getY());
      }
    } else {
      point.setX(ball.pos.getX());
      if (ball.pos.getY() < - 0.5 * PENALTY_AREA_WIDTH + 1) {
	point.setY( - 0.5 * PENALTY_AREA_WIDTH + 1);
      } else {
	point.setY( 0.5 * PENALTY_AREA_WIDTH - 1);
      }
    }
    Vector2d target = point - my.pos;
    
    dribbleDir = target.getAngle();
    
    bool dribbleSafe = ( snap->otherTeamInCorridor( my.pos, point, max( 10.0, 2 * target.getLength() ) ) == 0 &&
			 ( ( tsnap->firstDangerousOpponent == UNKNOWN ) ||
			   ( tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent ) > 3 ) ) );

    if ( dribbleSafe )
      g.set_GOOD();
    else
      g.set_BAD();

    fd_dashPower = MAXPOWER;   // assume goalie can dash at full speed
    fd_dist = target.getLength();

    double dist = 0;
    double speed = 0;

    // start with step 3 because 2 turns are the standard case here!
    for( fd_numSteps = 3; fd_numSteps < 20; fd_numSteps++ )
      {
	speed += fd_dashPower * snap->my.dashPowerRate;
	dist += speed;
	speed *= snap->my.decay;
	if( dist > fd_dist )
	  break;
      }

    if( fd_numSteps * ( MAXPOWER - snap->my.staminaIncMax ) > availableStamina || fd_numSteps > 12 )
      { 
	g.set_BAD();
	BHTRC(MLog("dribble -> cannot kick ball back into penalty area safely, stamina too low!"
		   " Need passmodule to do something useful !"));
      }
  }
  
  return g;
}


Grade 
BallHandlingModule::evaluatePlayer()
{
  Grade g;

  if (skills->kickSafe(snap)) {
    g = evaluateDribble(); // try to dribble, if possible
    return g;
  }

  // get ball if frp or pass receiver or controller
  if (((tsnap->ourAction.state == TA_PASS) &&
       (tsnap->ourAction.receiver == my.number)) ||
      ((tsnap->ourAction.state == TA_CONTROL) &&
       (tsnap->ourAction.performer == my.number)) ||
      (tsnap->firstResponsiblePlayer == my.number)) {
    if (time - lastKickTime <= 1) {
      if (opponentKicked) {
	if (ball.pos.getX() < -FIELD_LENGTH/2 + PENALTY_AREA_LENGTH) {
	  // assume kick onto goal, and get that ball
	  Vector2d icptPos = ourGoal.pos - ball.pos;
	  double stretch = norm(icptPos);
	  if (stretch < 1e-4)
	    stretch = 1;
	  icptPos.stretch(2*BALL_SPEED_MAX/stretch);
	  icptPos += ball.pos;
	  turtleCmd = skills->goToPositionTimed(snap, icptPos, skills->kickRange, availableStamina);
	  if (turtleCmd.state != AdvSkills::TFINISHED) {
	    action = PLAYER_INTERCEPT_ALT;
	    g.set_AVERAGE();
	    GRADETRC("eval -> 1. resp, icpt assumed goal kick", g);
	    return g;
	  }
	  else {
	    action = WAIT;
	    GRADETRC("eval -> 1. resp, wait for assumed goal kick", g);
	    g.set_POOR();
	    return g;
	  }
	}
	/*
	else {
	  // should be a opp pass. Ignore until we know better
	  action = WAIT;
	  GRADETRC("eval -> 1. resp, but opp just played pass", g);
	  g.set_BAD();
	  return g;
	}
	*/
      }
      else {
	// assume myself as pass receiver. We cannot check for TA_PASS
	// since this information cannot have reached us yet.
	// So wait for the ball. But if we feel the ball, we can intercept.
	// That is especially necessary since otherwise we would refuse
	// to dribble...
	if (distToBall > 2*BALL_SPEED_MAX) {
	  action = WAIT;
	  GRADETRC("eval -> 1. resp, wait for pass", g);
	  g.set_PERFECT();
	  return g;
	}
	// fall through to standard icpt
      }
    }

    action = PLAYER_INTERCEPT;
    if (tsnap->ourAction.state == TA_REACT) {
      if ((tsnap->firstDangerousOpponent != UNKNOWN) &&
	  (tsnap->ourTeamReachBallTime(tsnap->firstResponsiblePlayer) >
	   tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent)))
	g.set_AVERAGE();
      else
	g.set_GOOD();
    }
    else
      g.set_PERFECT();
    GRADETRC("eval -> 1. resp, icpt", g);
    return g;
  }

  // it only makes sense to provide fallback if we surely know where the ball rolls
  if ((time - lastKickTime > 1) &&
      ((tsnap->secondResponsiblePlayer == my.number) ||
       (tsnap->thirdResponsiblePlayer == my.number)) &&
      (tsnap->firstResponsiblePlayer != UNKNOWN)) {
    // see if I can go to a fallback position.
    int catchTime = (tsnap->secondResponsiblePlayer == my.number) ? 5 : 10;
    catchTime += tsnap->ourTeamReachBallTime(tsnap->firstResponsiblePlayer);
    Vector2d icptPos = ball.pos + ((1 - pow(BALL_DECAY, catchTime + 1))/(1 - BALL_DECAY))*ball.v;
    BHTRC(POSITION(icptPos));
    BHTRC(TEXT("icptPos", icptPos));
    
    turtleCmd = skills->goToPositionTimed(snap, icptPos, skills->kickRange, availableStamina);
    if (turtleCmd.state == AdvSkills::TFINISHED) {
      // seldom, but just wait for ball...
      g.set_GOOD();
      action = WAIT;
      GRADETRC("eval -> 2./3. resp, waiting", g);
      return g;
    }

    if (turtleCmd.time < catchTime) {
      g.set_GOOD();
      action = PLAYER_INTERCEPT_ALT;
      GRADETRC("eval -> 2./3. resp, icpt alt", g);
      return g;
    }
  }

  action = WAIT;
  g.set_BAD();
  return g;
}


const BallHandlingModule::DribbleAreaData
BallHandlingModule::areas[] = {
  // 1. right opponent corner-triangle
  {
    5, {
      { LEFT,           Grade::GOOD,    false },   // true means: no fastdribble
      { LEFT_LEFT_BACK, Grade::GOOD,    false },
      { LEFT_BACK,      Grade::AVERAGE, false },
      { BACK_LEFT_BACK, Grade::POOR,    false },
      { BACK,           Grade::POOR,    false }
    },
    Angle(-M_PI/2)
  },
  // 2. left opponent corner-triangle
  {
    5, {
      { RIGHT,            Grade::GOOD,    false },
      { RIGHT_RIGHT_BACK, Grade::GOOD,    false },
      { RIGHT_BACK,       Grade::AVERAGE, false },
      { BACK_RIGHT_BACK,  Grade::POOR,    false },
      {	BACK,             Grade::POOR,    false }
    },
    Angle(M_PI/2)
  },
  // 3. right border
  {
    3, {
      { LEFT_FORWARD,         Grade::GOOD,    false },
      { FORWARD_LEFT_FORWARD, Grade::GOOD,    false },
      { FORWARD,              Grade::AVERAGE, true  }
    },
    Angle(-M_PI/4)
  },
  // 4. left border
  {
    3, {
      { RIGHT_FORWARD,         Grade::GOOD,    false },
      { FORWARD_RIGHT_FORWARD, Grade::GOOD,    false },
      { FORWARD,               Grade::AVERAGE, true  }
    },
    Angle(M_PI/4)
  },
  // 5. our right backfield
  {
    5, {
      { RIGHT_FORWARD,         Grade::AVERAGE, false },
      { RIGHT_RIGHT_FORWARD,   Grade::AVERAGE, false },
      { RIGHT,                 Grade::AVERAGE, false },
      { FORWARD_RIGHT_FORWARD, Grade::POOR,    false },
      { FORWARD,               Grade::POOR,    false }
    },
    Angle(0.0)
  },
  // 6. our left backfield
  {
    5, {
      { LEFT_FORWARD,         Grade::AVERAGE, false },
      { LEFT_LEFT_FORWARD,    Grade::AVERAGE, false },
      { LEFT,                 Grade::AVERAGE, false },
      { FORWARD_LEFT_FORWARD, Grade::POOR,    false },
      { FORWARD,              Grade::POOR,    false }
    },
    Angle(0.0)
  },
  // 7. right front flank (Axel's "Leitplanke")
  {
    6, {
      { FORWARD,              Grade::GOOD,    false },
      { FORWARD_LEFT_FORWARD, Grade::PERFECT, false },
      { LEFT_FORWARD,         Grade::PERFECT, false },
      { LEFT_LEFT_FORWARD,    Grade::GOOD,    false },
      { LEFT,                 Grade::AVERAGE, false },
      { BACK,                 Grade::POOR,    false }
    },
    Angle(-M_PI/2)
  },
  // 8. left front flank
  {
    6, {
      { FORWARD,                Grade::GOOD,    false },
      { FORWARD_RIGHT_FORWARD,  Grade::PERFECT, false },
      { RIGHT_FORWARD,          Grade::PERFECT, false },
      { RIGHT_RIGHT_FORWARD,    Grade::GOOD,    false },
      { RIGHT,                  Grade::AVERAGE, false },
      { BACK,                   Grade::POOR,    false }
    },
    Angle(M_PI/2)    
  },
  // 9. somewhere else in the right body of the field
  {
    9, {
      { FORWARD,               Grade::PERFECT, false },
      { FORWARD_RIGHT_FORWARD, Grade::PERFECT, false },
      { RIGHT_FORWARD,         Grade::PERFECT, false },
      { RIGHT_RIGHT_FORWARD,   Grade::GOOD,    false },
      { FORWARD_LEFT_FORWARD,  Grade::GOOD,    false },
      { LEFT_FORWARD,          Grade::AVERAGE, false },
      { RIGHT,                 Grade::AVERAGE, true  },
      { LEFT,                  Grade::POOR,    false },
      { LEFT_LEFT_FORWARD,     Grade::POOR,    false }
    },
    Angle(M_PI/8)
  },
  // 10. somewhere else in the left body of the field
  {
    9, {
      { FORWARD,               Grade::PERFECT, false },
      { FORWARD_LEFT_FORWARD,  Grade::PERFECT, false },
      { LEFT_FORWARD,          Grade::PERFECT, false },
      { LEFT_LEFT_FORWARD,     Grade::GOOD,    false },
      { FORWARD_RIGHT_FORWARD, Grade::GOOD,    false },
      { RIGHT_FORWARD,         Grade::AVERAGE, false },
      { LEFT,                  Grade::AVERAGE, true  },
      { RIGHT_RIGHT_FORWARD,   Grade::POOR,    false },
      { RIGHT,                 Grade::POOR,    false }
    },
    Angle(-M_PI/8)
  }
};

int BallHandlingModule::dribbleArea()
{
  const double sideBorder(5.0);
  const double backBorder(1.5*PENALTY_AREA_LENGTH);
  const double cornerDist((FIELD_WIDTH - GOAL_WIDTH)/2);
  const double targetingDist(cornerDist + GOAL_WIDTH/2);

  // the dribble directions depend on the player's position on the field:
  // 1. right opponent corner-triangle
  if (1.5*(FIELD_LENGTH/2 - my.pos.getX()) +
      (FIELD_WIDTH/2 - my.pos.getY()) < cornerDist)
    return 0;
  // 2. left opponent corner-triangle
  if (1.5*(FIELD_LENGTH/2 - my.pos.getX()) +
      (FIELD_WIDTH/2 + my.pos.getY()) < cornerDist)
    return 1;
  // 3. right border
  if (my.pos.getY() > FIELD_WIDTH/2 - sideBorder)
    return 2;
  // 4. left border
  if (my.pos.getY() < -FIELD_WIDTH/2 + sideBorder)
    return 3;
  // 5./6. our backfield
  if (my.pos.getX() < -FIELD_LENGTH/2 + backBorder) {
    if (my.pos.getY() > 0)
      return 4;
    else
      return 5;
  }
  // 7. right front flank (Axel's "Leitplanke")
  if (1.0*(FIELD_LENGTH/2 - my.pos.getX()) +
	   (FIELD_WIDTH/2 - my.pos.getY()) < targetingDist)
    return 6;
  // 8. left front flank
  if (1.0*(FIELD_LENGTH/2 - my.pos.getX()) +
      (FIELD_WIDTH/2 + my.pos.getY()) < targetingDist)
    return 7;
  // 9./10. somewhere else in the body of the field
  if (my.pos.getY() > 0)
    return 8;
  else
    return 9;
}

bool BallHandlingModule::checkDribbleDir(Vector2d target, double secDist)
{
  if ((snap->otherTeamInCorridor(my.pos, 2*target + my.pos, 2*norm(target), PI_4) == 0) &&
      (inField(my.pos + target, secDist)))
    return true;
  return false;
}

Grade 
BallHandlingModule::evaluateDribble()
{
  Grade g;
  // first: if we do not know if a collision has happened, ban fast dribble
  // but only if more or less alone
  if ((snap->collState != NONE) &&
      ((tsnap->firstDangerousOpponent == UNKNOWN) ||
       (tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) >
	snap->otherTeam(tsnap->firstDangerousOpponent).age))) {
    BHTRC(MLog("dribble -> unsafe collision state"));
    action = PLAYER_PLACE;
    placeDir = (ball.pos - my.pos).getAngle();
    g.set_PERFECT();
    return g;
  }

  // minimum dist to field border watched while dribbling
  const double sideBorder(5.0);

  const DribbleAreaData *adt = &areas[dribbleArea()];
  double savedStamina = min (2*MAXPOWER, availableStamina);

  // keep back some stamina in case we lose the ball while dribbling, so we can
  // intercept the ball again within 2-3 steps.
  dribbleStamina = max( 0.0, availableStamina - savedStamina );
  int    nDashes = 0;
  double dist = 0.0, dashPower = 0.0;
  bool   opt_dribbleonly = true; // fastdribble/dribble switch
  Grade  opt_g(Grade::BAD);
  int    opt_dir = 0;
  int    opt_nDashes = 2;
  double opt_dashPower = 0;
  double opt_dist = 0;
  Vector2d opt_vec;

  for (int cdir = 0; cdir < adt->numDirs; cdir++) {
    Vector2d propReachVec;

    // check if fastdribble is allowed for that direction
    if (!adt->dirs[cdir].dribble) {
      for (nDashes = 20; nDashes >= 5; nDashes -= 3) {
	// check several fastdribble distances 
	// (iterating number of dashes needed for intercept)
	dashPower = snap->my.dashPowerRate * min( MAXPOWER, snap->my.staminaIncMax + dribbleStamina / nDashes );

	dist = 0.0;
	double speed = 0.0;

	for( int i = 0; i < nDashes; i++ )
	  {
	    speed += dashPower;
	    dist += speed;
	    speed *= snap->my.decay;
	  }

	propReachVec = dist * adt->dirs[cdir].dir;

	if (checkDribbleDir(propReachVec, sideBorder)) {
	  Grade g = adt->dirs[cdir].grade;
	  if(g.is_better_than(opt_g)){
	    opt_dribbleonly = false;
	    opt_g           = g;
	    opt_dir         = cdir;
	    opt_dashPower   = dashPower;
	    opt_nDashes     = nDashes;
	    opt_dist        = dist;
	    opt_vec         = propReachVec;
	    break;
	  }
	}
      }
    }

    // check simple dribble
    propReachVec = STAMINA_INC_MAX*my.dashPowerRate*adt->dirs[cdir].dir;
    double ballDistToBorder = min( fabs( snap->ball.pos.getX() ) - FIELD_LENGTH/2.0,
				   fabs( snap->ball.pos.getY() ) - FIELD_WIDTH/2.0 ); 
    // Dribble should take ball away from the border
    if (checkDribbleDir(propReachVec, ballDistToBorder)) {
      Grade g = adt->dirs[cdir].grade;
      // g.downgrade();   // why that? (CM)
      if(g.is_better_than(opt_g)){
	opt_dribbleonly = true;
	opt_g   = g;
	opt_dir = cdir;
	opt_vec = propReachVec;
      }
    }
  }

  if(!opt_g.is_BAD()) {
    // we have got a very nice and handsome dribble direction
    g = opt_g;
    dribbleDir = adt->dirs[opt_dir].dir.getAngle();

    if (fabs(dribbleDir - my.faceDir) > M_PI/4.0) {
      opt_dribbleonly = true;
      g.downgrade();
      BHTRC(MLog("dribble -> delayed fd to th"));
    }

    action = opt_dribbleonly ? PLAYER_DRIBBLE : PLAYER_FAST_DRIBBLE;

    if (opt_dribbleonly) {
      BHTRC(MLog("dribble -> dribble slow"));
      BHTRC(CORRIDOR(my.pos, 2*opt_vec + my.pos, 2*norm(opt_vec), PI_4));
    }
    else {
      fd_dashPower = opt_dashPower;
      // spare two cycles for oddities (bodyDir!), remark: addition here! (CM)
      fd_numSteps  = opt_nDashes + 2;
      fd_dist      = opt_dist;

      BHTRC(CORRIDOR(my.pos, 2*opt_vec + my.pos, 2*norm(opt_vec), PI_4));
#ifdef BALLHTRACE
      ostrstream tmp;
      tmp << "dribble -> fast " << dribbleDir << " " << norm(opt_vec) << " "
	  << fd_numSteps << " " << fd_dashPower << (char)0;
      BHTRC(MLog(tmp.str()));
#endif
    }
    
    // direction where to place ball when no opponent in sight
    // (if near a border, place to inside of field)
    double defaultdir = dribbleDir + (((ball.pos - my.pos).getAngle() - dribbleDir > 0) ? PI_3 : -PI_3);

    /* (CM)
    if (my.pos.getY() > FIELD_WIDTH/2 - sideBorder)
      defaultdir = -PI_2;
    else if (my.pos.getY() < -FIELD_WIDTH/2 + sideBorder)
      defaultdir = PI_2;
    else if (my.pos.getX() > FIELD_LENGTH/2 - sideBorder)
      defaultdir = PI;
    else if (my.pos.getX() < -FIELD_LENGTH/2 + sideBorder)
      defaultdir = 0;
    */

    // place to defaultdir if firstDangerous is far away
    if ((tsnap->firstDangerousOpponent == UNKNOWN) ||
	(my.distTo(snap->otherTeam(tsnap->firstDangerousOpponent)) > 8.0*PLAYER_SIZE))
      holdDir = defaultdir;
    else {
      // also place to defaultdir if secondDangerous is also really dangerous
      // [!] place dir if we have two dangerous opponents is not ok
      //if ((tsnap->secondDangerousOpponent != UNKNOWN) &&
      //	  (my.distTo(snap->otherTeam(tsnap->secondDangerousOpponent)) < 8.0*PLAYER_SIZE))
      //	holdDir = defaultdir;
      //else {
      // place away from first dangerous opponent
      Angle opponentDir = (snap->otherTeam(tsnap->firstDangerousOpponent).pos - my.pos).getAngle();
      holdDir = dribbleDir + ((opponentDir - dribbleDir > 0) ? -PI_3 : PI_3);
      // }
    }
    BHTRC(LINE(my.pos, my.pos + radAng2Vector2d(1, holdDir)));
    BHTRC(TEXT("H", my.pos + radAng2Vector2d(1, holdDir)));
    return g;
  }

  // no direction for dribbling found
  action = PLAYER_PLACE;

  // place somewhere if firstDangerous is far away
  // [!] placing should not just be opposite side, but 
  // could even try to trick opponent (place to one orthogonal side of my->mytargetdir vector
  // while opp is medium far away, when he gets too close, move ball around back to make him
  // pass us on one side, and way is free...)

  if ((tsnap->firstDangerousOpponent == UNKNOWN) ||
      (my.distTo(snap->otherTeam(tsnap->firstDangerousOpponent)) 
       > 8.0*PLAYER_SIZE)) {
    placeDir = (ball.pos - my.pos).getAngle();
    // opponent far away, therefore just take some breath and keep ball where it is.
  } else {
    if ((tsnap->secondDangerousOpponent != UNKNOWN) &&
	(my.distTo(snap->otherTeam(tsnap->secondDangerousOpponent)) < 8.0*PLAYER_SIZE)) {
      placeDir = (ball.pos - my.pos).getAngle();
      // [!] find a better place pos if two opponents are close
    } else {
      Angle opponentDir = (snap->otherTeam(tsnap->firstDangerousOpponent).pos - my.pos).getAngle();
      // opposite side of closest opponent
      holdDir = PI + opponentDir;
    }
  }

  g.set_BAD();
  return g;
}


Grade
BallHandlingModule::evaluateFindBall()
{
  Grade physTurnGrade;   // turn body?

  // turnHeadGrade is changed in this method; the result is used in turnHeadEvaluate()
  turnHeadGrade.set_BAD();
  physTurnGrade.set_BAD();   // init with BAD

  reqWidth = VIEW_WIDTH_NORMAL;
  
  // Idea: Both turnHeadGrade and physTurnGrade are upgraded as information about the
  // ball gets worse. If physTurnGrade is high enough findBall() should be called by the
  // mastercontrol. In this case the head is turned into player's body direction.
  // But if findBall() is not called the head is turned towards the ball if the
  // mastercontrol lets ballhandling do the turn head command.

  // --------------------< step 1 >--------------------
  // check how much time ball has spent in view cone

  int maxSeeGap;
  switch (snap->senseBody.viewWidth)
    {
    case VIEW_WIDTH_NARROW:
      maxSeeGap = 2;
      break;
    case VIEW_WIDTH_NORMAL:
      maxSeeGap = 4;
      break;
    case VIEW_WIDTH_WIDE:
      maxSeeGap = 6;
      break;
    default:   // not to happen
      maxSeeGap = 10;
    }

  // Axel's debug output for coneTime
#ifdef BALLHTRACE
  ostrstream tmp;
  tmp << "find -> coneTime " << snap->ball.coneTime << (char) 0;
  BHTRC(MLog(tmp.str()));
#endif

  if( snap->ball.coneTime >= maxSeeGap )
    {
      // should have seen ball (because position is in my view cone) but did not see it
      turnHeadGrade.set_PERFECT();   // -> turn head first
    }

  if( snap->ball.coneTime >= maxSeeGap + 2 )
    {
      // should have seen ball (because position is in my view cone) but did not see it
      physTurnGrade.set_PERFECT();   // -> turn body
      turnHeadGrade.set_PERFECT();   // -> turning body but not head does not make sense
    }
 
  // --------------------< step 2 >--------------------
  // check ball age; if it is too old, upgrade body and head turn

  int acceptableAge = 10;
  if( snap->my.number == tsnap->firstResponsiblePlayer )
    acceptableAge = maxSeeGap;
  else if( snap->my.number == tsnap->secondResponsiblePlayer )
    acceptableAge = ( 3 * maxSeeGap ) / 2;

  int diff = snap->ball.age - acceptableAge;
  // for each delay make turn more attractive
  if (diff > 0)
    {
      // physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
    }
  if (diff > 1)
    {
      physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
    }
  if (diff > 3)
    {
      physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
    }
  if (diff > 5)
    {
      physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
    }
  if (diff > 7)
    {
      physTurnGrade.upgrade();
      turnHeadGrade.upgrade();
    }

  // --------------------< step 3 >--------------------
  // check if we are close to the ball and someone could have
  // kicked it. In this case we should turn our head towards the
  // ball. This has not much to do with old data - in contrary!

  // in near range, look faster, since it is likely that we are not totally
  // wrong with the direction of the ball
  if( distToBall < 25.0 * PLAYER_SPEED_MAX )
    reqWidth = VIEW_WIDTH_NARROW;

  // check whether a player can kick
  // but if it is me, calm down. I should know what I do
  if( ( time - lastKickTime <= 1 ) &&
     ( distToBall < 25.0 * PLAYER_SPEED_MAX ) &&
      ( distToBall > 2.0 * BALL_SPEED_MAX ) )
    turnHeadGrade.set_PERFECT();

  GRADETRC("find -> current turn head grade : ", turnHeadGrade);
  GRADETRC("find -> current body turn grade : ", physTurnGrade);

  return physTurnGrade;
}
