//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player class goalshoteval
//
// \$Id: goalshoteval.h,v 1.9 2001/08/04 02:56:59 mrb Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////


#ifndef __GOALSHOTEVAL_H__
#define __GOALSHOTEVAL_H__

#include "grade.h"
#include "drawMacros.h"
#include "mcommand.h"
#include "worldmodel.h"
#include "Config.h"

#include <strstream.h>



// enum-type for our team / the other team
//////////////////////////////////////////////

enum team_type{
  OUR_TEAM,
  OTHER_TEAM
};


// the class GoalShotEval
///////////////////////////////

class GoalShotEval
{
public:
  // constructor of module GoalShotEval 
  GoalShotEval();

  // destructor of module GoalShotEval 
  ~GoalShotEval();
  
  // do the evaluation for a player of our team
  // for definition of Grade: see grade.h
  Grade ourEval(const Snapshot* worldSnap, Vector2d playerpos) const;

  // do the evaluation for a player of the other team
  // for definition of Grade: see grade.h
  Grade otherEval(const Snapshot* worldSnap, Vector2d playerpos) const;


private:
 
  // some constants this class needs
  //////////////////////////////////////

  // important target positions for our team in the others goal
  Vector2d leftEdgeOfGoal;
  Vector2d middleOfGoal;
  Vector2d rightEdgeOfGoal;

  // important targetpositions for the other team in our goal
  Vector2d leftEdgeOfOurGoal;
  Vector2d rightEdgeOfOurGoal;
  Vector2d middleOfOurGoal;

  // distances to the edges of the goal where the player should kick
  // to - depends on the distance between the player and the goal
  double minDistToCorner;
  double maxDistToCorner;

  // maximal time to do a kick
  int maxKickDuration;

  // distance a ball can be maximaly kicked
  double maxKickLength;

  // the lower bound for the distance when a goalshot will get 
  // distancegrade IMPOSSIBLE
  double distGradeIsImpDist;

  // the lower bound for the distance when a goalshot will get 
  // distancegrade BAD
  double distGradeIsBadDist;

  // the lower bound for the distance when a goalshot will get 
  // distancegrade POOR
  double distGradeIsPoorDist;

  // the lower bound for the distance when a goalshot will get 
  // distancegrade AVERAGE
  double distGradeIsAverageDist;

  // the lower bound for the distance when a goalshot will get 
  // distancegrade GOOD
  double distGradeIsGoodDist;

  // the distance when a goalshot will get garde PERFECT in any 
  // case
  double anyCaseGoalShotDist;

  // the distance to the goalie when a player decides to kick 
  // anyway
  double goalieNearDist;

  // if the player is outside the penalty area, his x-position
  // should at least have the following distance to the goal line
  // for that the angle towards the goal does not become too small
  double xDistToGoalLine;

  // the opening angle of the corridor from the player to the selected
  // target
  Angle corridorAngle;

  // the width of the corridor from the player to the selected
  // target where the other's goalie shall not be in
  double corridorWidth;

  // the width the corridor shall have when the corridor with
  // the above width ist not free
  double corridorWidthSmall;
  
  // elder things will not be considered
  int maxAge;   

  // next to the goalline (SECURITY_DIST in goalshoteval.C) a 
  // goalshot is considered impossible, if farer away from goal than
  double goalLineImpDist;

  // some private methods
  /////////////////////////////////////////

  // calculates the players minimum distance to the opponents goal,
  // considering the left and right edge as well as
  // the middle of the goal; returns the minimum distance
  // and writes the belonging postion in the goal into
  // target; manioulates also the corners of the goal depending 
  // on the distance of the player to the goal
  void targetInGoal(Vector2d playerpos, 
		    Vector2d &minTarget, 
		    Vector2d &leftTarget,
		    Vector2d &middleTarget, 
		    Vector2d &rightTarget,
		    double &minDist) const;

  // calculates a correction dist to the corner of the goal where the
  // player finaly should kick to
  Vector2d correctCorner(double dist) const;

  // answer the question, how much time a player has for doing a kick
  // before an opponent can reach the ball
  int availableTimeForKick(const Snapshot *worldSnap, Vector2d playerpos, 
			   team_type team = OUR_TEAM) const;

  // calculates the distance of an object to the straight line between
  // source and target
  // if the projection of the object onto the line source->target
  // is not a point on this line the function returns either
  // the dist to the source or the dist to target
  double distToLine(Vector2d object, Vector2d source, Vector2d target) const;  


  // evaluates the given corridor
  Grade evaluateCorridor(const Snapshot *worldSnap, Vector2d playerpos, 
			 Vector2d target, 
			 Angle angle, Grade startGrade, 
			 team_type team = OUR_TEAM) const;

  // determines the best corridor for our team and evaluates it; 
  // returns the grade for the best corridor and the belonging target
  Grade baseSearchAndEvalForCorridor(const Snapshot *worldSnap, 
				     Vector2d playerpos,
				     Vector2d &target, 
				     Vector2d leftTarget,
				     Vector2d middleTarget,
				     Vector2d rightTarget,
				     bool useLeft, 
				     bool useMiddle,
				     bool useRight) const;

  // determines the best corridor for our team and evaluates it using 
  // baseSearchAndEvalForCorridor and some more criteria; 
  // returns the grade for the best corridor and the belonging target
  Grade searchAndEvalOurBestCorridor(const Snapshot *worldSnap, 
				     Vector2d playerpos,
				     Vector2d &target,
				     Vector2d leftTarget,
				     Vector2d middleTarget,
				     Vector2d rightTarget) const;

  // determines the best corridor for the other team and evaluates it; 
  // returns the grade for the best corridor
  Grade searchAndEvalOtherBestCorridor(const Snapshot *worldSnap, 
				       Vector2d playerpos,
				       Vector2d &target,
				       Vector2d leftTarget,
				       Vector2d middleTarget,
				       Vector2d rightTarget) const;

  friend class GoalShotModule;    // this module will use some
                                  // of the private methods
 
}; // class GoalShotEval



#endif

