//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

///////////////////////////
//
// class PassEval
//
///////////////////////////

#include "passeval.h"
#include "Grid.h"


#undef USEPAEVTRACE
#undef USETESTPAEVTRACE
#undef USEPASSDRAWTRACE

// comment out the following line, if trace shall be disabled
//  #define USEPAEVTRACE
//  #define USETESTPAEVTRACE
//  #define USEPASSDRAWTRACE

#ifdef USEPAEVTRACE
#define PAEVTRACE(cmd) cmd
#warning PassEval debug trace PAEVTRACE is activated!
#else
#define PAEVTRACE(cmd)
#endif


#ifdef USETESTPAEVTRACE
#define TESTPAEVTRACE(cmd) cmd
#warning PassEval debug trace TESTPAEVTRACE is activated!
#else
#define TESTPAEVTRACE(cmd)
#endif

#ifdef USEPASSDRAWTRACE
#define PASSDRAWTRACE(cmd) cmd
#warning PassEval debug trace PASSDRAWTRACE is activated!
#else
#define PASSDRAWTRACE(cmd)
#endif


////////////////////////////
//
// Constructor
//
///////////////////////////

PassEval::PassEval(const GoalShotEval * _goalshoteval):
  goalshoteval(_goalshoteval)
{
  DANGERZONE          = Config::getValueAsDouble("pass_dangerzone");
  SAMELEVEL_LENGTH    = Config::getValueAsDouble("pass_sameLevelLength");
  MAXKICK_DURATION    = Config::getValueAsInt("pass_maxKickDuration");
  STDSITKICK_DURATION = Config::getValueAsInt("pass_stdSitKickDuration");
  MAXPASS_LENGTH      = Config::getValueAsDouble("pass_maxKickLength");
  RPASS_LATTICELENGTH = Config::getValueAsDouble("pass_rpassLatticeLength");
  SHORTPASS_LENGTH    = Config::getValueAsDouble("pass_shortPassLength");
  SHORTPASS_LENGTH_FOR_GOALIE = Config::getValueAsDouble("pass_shortPassLengthForGoalie");
  SECURITY_DIST       = Config::getValueAsDouble("pass_securityDist");
  SIDELINE_DIST       = Config::getValueAsDouble("pass_sideLineDist");

  goalevalpoints = new GoalEvalPoints();
}

////////////////////////////
//
// Destructor
//
///////////////////////////

PassEval::~PassEval()
{}  

//////////////////////////////////////////////////////////////////////////
//
// public methods
//
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
//
// Grade PassEval::ourEval - with rangepasses
//
/////////////////////////////////////////////////

Grade PassEval::ourEval(const Snapshot* worldSnap, int plnumber, 
			int panumber, Vector2d &posOfPass) const
{
  PropOf pprop;
  get_pPass(true, worldSnap, plnumber, worldSnap->ourTeam(panumber).pos, 
	    panumber, pprop,
	    (worldSnap->ourTeam(plnumber).isGoalie == GOALIE_TRUE )? true:false);
  
  PropOf rprop;
  get_rPass(true, worldSnap, plnumber, panumber, rprop,
	   (worldSnap->ourTeam(plnumber).isGoalie == GOALIE_TRUE )? true:false );

  // we have only a p-pass
  if ( (pprop.perhaps) && (!rprop.perhaps) ) {
    posOfPass = pprop.pos;
    return pprop.grade;
  }
  // we have only a r-pass
  if ( (!pprop.perhaps) && (rprop.perhaps) ) {
    posOfPass = rprop.pos;
    return rprop.grade;
  }
  // we have both possibilities
  if ( (pprop.perhaps) && (rprop.perhaps) ) 
    if ( betterPass(worldSnap, rprop, pprop) ) {
      posOfPass = rprop.pos;
      return rprop.grade;
    }
    else {
      posOfPass = pprop.pos;
      return pprop.grade;
    }
  
  // there are no passes
  pprop.grade.set_BAD();
  return pprop.grade;
}

////////////////////////////
//
// Grade PassEval::ourEvalProp
//
///////////////////////////

PropOf PassEval::ourEvalProp(const Snapshot* worldSnap, int plnumber, 
			     int panumber) const
{
  PropOf pprop;
  get_pPass(true, worldSnap, plnumber, worldSnap->ourTeam(panumber).pos, 
	    panumber, pprop,
	    (worldSnap->ourTeam(plnumber).isGoalie == GOALIE_TRUE )? true:false);
  
  PropOf rprop;
  get_rPass(true, worldSnap, plnumber, panumber, rprop,
	    (worldSnap->ourTeam(plnumber).isGoalie == GOALIE_TRUE )? true:false);

  // we have only a p-pass
  if ( (pprop.perhaps) && (!rprop.perhaps) )
    return pprop;

  // we have only a r-pass
  if ( (!pprop.perhaps) && (rprop.perhaps) )
    return rprop;

  // we have both possibilities
  if ( (pprop.perhaps) && (rprop.perhaps) ) 
    if ( betterPass(worldSnap, rprop, pprop) )
      return rprop;
    else
      return pprop;
  
  // there are no passes
  pprop.perhaps = false;
  pprop.grade.set_BAD();
  return pprop;
}

/////////////////////////////////
//
// Grade PassEval::otherEval
//
////////////////////////////////

Grade PassEval::otherEval(const Snapshot* worldSnap, int plnumber, 
			  int panumber) const
{
  PropOf prop;
  get_pPass(false, worldSnap, plnumber, worldSnap->otherTeam(panumber).pos, 
	    panumber, prop, false);
  
  PAEVTRACE({ostrstream ost;
  ost << " PASS: ourEval - got some props for other team:" 
      << prop << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  });
 
  return prop.grade;
}

////////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////////

////////////////////////////////////
//
// PassEval::check_for_safekick
//
///////////////////////////////////

int PassEval::check_for_safekick(const Snapshot *wS, int player) const
{ 
  
  // in these playmodes we have got enough time
  // just check if we must not place because ball is at field border
  if ( (wS->playMode == PLAY_MODE_GOAL_KICK_US) ||
       (wS->playMode == PLAY_MODE_FREE_KICK_US) ||
       (wS->playMode == PLAY_MODE_KICK_IN_US) ||
       (wS->playMode == PLAY_MODE_CORNER_KICK_US) )
    return ( inField(wS->ball.pos, SECURITY_DIST) ? 
	     MAXKICK_DURATION : STDSITKICK_DURATION );

  // closest opponent to player
  int opp = wS->otherTeamClosestTo(wS->ourTeam(player).pos);
  
  if (opp == UNKNOWN)          
    // player does not see any opponent so everything seems to be allright
    return ( inField(wS->ourTeam(player).pos, SECURITY_DIST) ? 
	     MAXKICK_DURATION : STDSITKICK_DURATION );
  
  // distance between player and closest opponent
  double pod = max( 0.0, dist( wS->ourTeam(player).pos, wS->otherTeam(opp).pos )
		   - wS->otherTeam(opp).kickRange() ); 
  
  // approximated time opponent will need to reach the ball
  // value should be at least 1 (what will the skills do else?)
  return max( STDSITKICK_DURATION, int( pod / wS->otherTeam(opp).vmax ) );
}


//////////////////////////////////
//
// bool PassEval::get_pPass
//
//////////////////////////////////

bool PassEval::get_pPass(bool we, const Snapshot *wS, int plnumber,
			 const Vector2d target, int panumber, 
			 PropOf &prop, bool isGoalie) const
{
  // [C] MAXAGE?
  double offsideLine = we ? 
    wS->ourOffsideLine(MAXAGE) : wS->otherOffsideLine(MAXAGE);

  Vector2d player    = we ? 
    wS->ourTeam(plnumber).pos : wS->otherTeam(plnumber).pos;

  Vector2d partner   = we ? 
    wS->ourTeam(panumber).pos : wS->otherTeam(panumber).pos;

 
  // [C] SUPERKICK_DURATION?
  int time_for_kick  = we ? 
    check_for_safekick(wS, plnumber) : MAXKICK_DURATION;

  // set the PropOf
  prop.grade.set_BAD();
  prop.perhaps = true;
  prop.number  = panumber;
  prop.pos     = target;
  prop.dist    = dist(player, target);

  // partner close and no dangerous situation
  if (time_for_kick > 1 )
    {
      if ( ( isGoalie && prop.dist < SHORTPASS_LENGTH_FOR_GOALIE )
	   || ( !isGoalie && prop.dist < SHORTPASS_LENGTH ) ) 
	{
	  prop.perhaps = false;
	  prop.grade.set_BAD();
	  return false;
	}
    }

  // partner too far away
  if ( MAXPASS_LENGTH < prop.dist ) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  prop.offside = ( we ? (offsideLine <= partner.getX()) : (offsideLine >= partner.getX()) );

  // if player is not offside, test teammates in a 10-m-corridor
  // if one of them is offside, do not pass
  if (!prop.offside) {
    for (int i=1; i <= 11; i++)
      if (i != plnumber && i != panumber)
	if (we) {
          // [C] Konstanten in inCorridor-Aufrufen
	  if ( (wS->ourTeam(i).inCorridor(player, target, 10, PI_3)) && 
	       (wS->ourTeam(i).pos.getX() >= offsideLine) )
	    prop.offside = true;
	}
	else {
	  if ( (wS->otherTeam(i).inCorridor(player, target, 10, PI_3)) && 
	       (wS->otherTeam(i).pos.getX() <= offsideLine) )
	    prop.offside = true;
	}
  }
  if (prop.offside) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  // estimated time player will need to reach the target position 
  // (good estimation for n large)

  const WorldPlayer &pl = ( we ? wS->ourTeam(plnumber) : wS->otherTeam(plnumber) );

  int time_to_reach = int( max( 0.0, dist(partner, target) - pl.kickRange() )
			   * (1 - pl.decay) / pl.staminaIncMax / pl.dashPowerRate );

  // calculate maximal velocity the ball may have at the beginning so 
  // that partner can reach it
  // [C] Konstanten, die den Ball betreffen
  // time_to_reach + 3 statt time_to_reach + 1, damit sicherheitshalber
  // 2 Zeitschritte zum Drehen mit eingerechnet werden
  prop.vel = min( BALL_SPEED_MAX, prop.dist * (1 - BALL_DECAY) / 
		  (1 - power(BALL_DECAY, time_to_reach + 3)) );

  // [C] Konstanten, die die POWER betreffen - und ich uebertreff' mich 
  // hier selber, mit C's!
  if (prop.vel < MAXPOWER/2 * KICK_POWER_RATE) {      
    // partner will not reach the pass
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  // set velocity to target velocity
  // [C] BALL_DECAY
  prop.vel *= power( BALL_DECAY, time_to_reach );

  // calculate grade and fill rest of prop
  prop.grade = calculateGrade(we, wS, player, target, prop, isGoalie);

  // partner not standing free or no pass corridor found
  if (prop.free < DANGERZONE || prop.corridor <= 0) {
    prop.perhaps = false;
    prop.grade.set_BAD();
    return false;
  }

  return true;
} // get_pPass

///////////////////////////////
//
// bool PassEval::get_rPass
//
///////////////////////////////

bool PassEval::get_rPass(bool we, const Snapshot *wS, int plnumber, 
			 int panumber, PropOf &prop, bool isGoalie) const
{
  // let be an array with:
  //
  //
  //  *  *  *  +       |       p = partner 
  //
  //  *  *  *  +       |       * = possible points to do a rangepass
  // 
  //  p  *  *  *       |       + = unwanted points
  //
  //  *  *  *  +      \ /      distance beetween points: 3 meters
  //
  //  *  *  *  +       j
  //
  //
  //  -  -  -  > i  

  // and here's the Grid:
  Grid <PropOf> rpropGrid(5, 4);
  PropOf helpProp;

  Vector2d player  = we ? 
    wS->ourTeam(plnumber).pos : wS->otherTeam(plnumber).pos;
  Vector2d partner = we ? 
    wS->ourTeam(panumber).pos : wS->otherTeam(panumber).pos;

  for (int i = 0; i < 4; i++)      // do something for all possible points
    for (int j = 0; j < 5; j++) {  // do it with rangepasses[i][j]
      // calculate the positions of the points
      helpProp.pos.setX(partner.getX() +     i * RPASS_LATTICELENGTH);
      helpProp.pos.setY(partner.getY() + (j-2) * RPASS_LATTICELENGTH);
      rpropGrid.set(j,i, &helpProp);

      // all rangepasses are a priori possible ...
      helpProp.perhaps = true;
      // ...but not the partner himself and not the ones marked with "+"
      if ( rpropGrid.get(j,i) &&  // 
           ( ((i == 0) && (j == 2)) || ((i == 3) && (j != 2)) || 
	     !inField(rpropGrid.get(j, i)->pos, SIDELINE_DIST) ) ) {
	helpProp.perhaps = false;
	helpProp.grade.set_BAD();
      }
      else {
	// if this rangepass is possible - we did by now nothing else 
	// than considering this question - 
	// calculate the rest of its propOf
	get_pPass(we, wS, plnumber, helpProp.pos, panumber, helpProp, isGoalie);

	if (helpProp.perhaps)
	  // is there an opponent that is closer to the target than 
	  // the partner?
	  if (dist(partner, helpProp.pos) > helpProp.free)
	    {
	      helpProp.perhaps = false;
	      helpProp.grade.set_BAD();
	    }
      } // else

      // put the helpProp to the rpropGrid
      rpropGrid.set(j, i, &helpProp);
      
      if (helpProp.perhaps)
        PASSDRAWTRACE(Comm::sendMCmd(MDPosition(helpProp.pos)));
    }

  TESTPAEVTRACE({ostrstream ost;
  ost << " checked r-passes for player" 
      << panumber << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  for (int jj = 0; jj < 5; jj++) {
    for (int ii = 0; ii < 4; ii++)
      {
	if (rpropGrid.get(jj, ii))
	  ost << (rpropGrid.get(jj, ii)->perhaps ? "1" : "0");
	else
	  ost << "-" ;
	Comm::sendMCmd(MLog(ost.str()));
      }
  }
  } );
  
  bool found_pass = false;

  // set prop to default values
  prop.grade.set_BAD();
  prop.perhaps = false;
  prop.number = -1;
  prop.offside = false;
  prop.corridor = 0;
  prop.free = 0;
  prop.dist = 0;
  prop.pos = player;
  prop.vel = 0;

  // now search the "best" rangepass
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 5; j++)
      if (rpropGrid.get(j,i) && rpropGrid.get(j, i)->perhaps)
	if (!found_pass) {
	  prop = *rpropGrid.get(j, i);
	  found_pass = true;
	}
	else if (betterPass(wS, *rpropGrid.get(j, i), prop))
	  prop = *rpropGrid.get(j, i);

  PAEVTRACE({ostrstream ost;
  ost << " check r-passes for player "
      << panumber << ", result: " << prop.grade
      << char(0);
  Comm::sendMCmd(MLog(ost.str()));
  } );
  
  return found_pass;
} // get_rPass



///////////////////////////////
//
// PassEval::calculateGrade
//
//////////////////////////////

Grade PassEval::calculateGrade(bool we, 
			       const Snapshot *wS, const Vector2d &player,
			       const Vector2d &partner,
			       PropOf &prop, bool isGoalie) const
{
  double playerx  = player.getX();
  double playery  = player.getY(); 
  double partnerx = partner.getX();
  double partnery = partner.getY();

  WorldObject ourGoal   = ( we ? wS->ourGoal : wS->otherGoal );
  WorldObject otherGoal = ( we ? wS->otherGoal : wS->ourGoal );

  // calculate maximal corridor width and maximal free circle around partner
  // [C] 20, 4*4? he, das ist nicht von mir! Baeggi!
  prop.corridor = 20;
  prop.free = 20;

  if (we) {
    int closestOpponent = wS->otherTeamClosestTo(partner);
    if (closestOpponent != UNKNOWN)
      prop.free = int(dist(partner,wS->otherTeam(closestOpponent).pos));
    prop.corridor = min(20, prop.free/4*4);
    // [C] Korridorwinkel
    while (wS->otherTeamInCorridor(player, partner, prop.corridor, PI_6) 
	   && prop.corridor > 0) {
      prop.corridor -= 4;
    }
  }
  else {
    int closestOpponent = wS->ourTeamClosestTo(partner);
    if (closestOpponent != UNKNOWN)
      prop.free = int(dist(partner,wS->ourTeam(closestOpponent).pos));
    prop.corridor = min(20, prop.free/4*4);
    // [C] nochmal eine Winkelkonstante
    while (wS->ourTeamInCorridor(player, partner, prop.corridor, PI_6) 
	   && prop.corridor > 0) {
      prop.corridor -= 4;
    }
  }

  // [C] das sind wohl auch Erfahrungswerte, oder? 
  // und das in der ganzen Methode!!!
  switch (prop.corridor) {
  case 20 :
    prop.grade.set_PERFECT(); break;
  case 16 :
    prop.grade.set_PERFECT(); break;
  case 12 :
    prop.grade.set_PERFECT(); break;
  case  8 :
    prop.grade.set_PERFECT();
    if (prop.dist > 20)         // better replace that by a constant
      prop.grade.downgrade();
    break;
  case  4 :
    prop.grade.set_PERFECT();
    if (prop.dist > 20)
      prop.grade.downgrade();
    if (prop.dist > 16)
      prop.grade.downgrade();
    if (prop.dist > 12)
      prop.grade.downgrade();
    break;
  case  0 :
    prop.grade.set_BAD(); return prop.grade;
  default :
    prop.grade.set_BAD(); return prop.grade;   // should not happen
  }

  // if this method is called for the other team, 
  // swap sign of the x-coordinates
  // so that the following algorithm will work
  if (!we) {
    playerx  = -playerx;
    partnerx = -partnerx;
  }

  // partner in own third (-> defensive situation)
  // [C] es folgen ganz viele Konstanten 
  if (partnerx < -FIELD_LENGTH/6) {
    // pass should take the ball away from goal
    if (sqrnorm(partner - ourGoal.pos) < sqrnorm(player - ourGoal.pos)) {
      prop.grade.downgrade();
    }
    // pass should not cross the line y=0 if the pass is too short
    // AENDERUNG: bringt der Pass einen grossen x-Gewinn, dann  darf er auch ueber
    // die Mittellinie gehen
    if ( (fabs(playery) > 5)
	 && (playery * partnery < 0) 
	 && (partnerx - playerx < 15)) {
      prop.grade.downgrade();
    }
    // no passing back
    // AENDERUNG : 2* satt 1*
    if (partnerx - playerx < 2*SAMELEVEL_LENGTH) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -8) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -15) {
      prop.grade.downgrade();
    }
    if (partnerx - playerx < -20) {
      prop.grade.downgrade();
    }
  }  
  // partner in middle third (-> midfield situation)
  else if (partnerx < FIELD_LENGTH/4) {
    // pass should not cross the line y=0 if player is too far from it
    // AENDERUNG: bringt der Pass genug x-Gewinn, ist er ok
    if ( (fabs(playery) > 10) 
	 && (playery * partnery < 0)
	 && (partnerx < 15 + playerx  )) 
      {    
	prop.grade.downgrade();
      }

    // pass towards the middle of the field
    if ( fabs(partnery) < fabs(playery) - 10 )
      {
	prop.grade.downgrade();
      }

   // AENDERUNG: 2-faches downgrade eines Passes zur Mitte, 
   // falls zu geringer x-Gewinn
    if ( ( fabs(partnery) < fabs(playery) - 10)
         && (partnerx < playerx + 15 ) )
      prop.grade.downgrade();

    // partner behind
    if ( (partnerx < playerx + SAMELEVEL_LENGTH)) 
      {
	prop.grade.downgrade();
      }
    
    // AENDERUNG: falls Partner hintendran und kein y-Gewinn, weiteres downgrade
    if( (partnerx < playerx + SAMELEVEL_LENGTH)
	&& ( fabs(partnery) < fabs(playery) + 12 ))
      prop.grade.downgrade();

    if (partnerx < playerx - 10) {
      prop.grade.downgrade();
    }
    if (partnerx < playerx - 15) {
      prop.grade.downgrade();
    }
    if (partnerx < playerx - 20) {
      prop.grade.downgrade();
    }
    // if player himself is far in front, downgrade
    // (this prevents passes back into midfield)
    if( playerx > FIELD_LENGTH / 4.0 ) {
      prop.grade.downgrade();
    }

  }
  // partner in opponent's quarter (->offensive situation)  
  else {
    // partner too near to the border and too far in front
    if ( (1.5*(FIELD_LENGTH/2.0 - partnerx) +
	  (FIELD_WIDTH/2.0 - partnery) < (FIELD_WIDTH - GOAL_WIDTH)/2.0) ||
	 (1.5*(FIELD_LENGTH/2.0 - partnerx) +
	  (FIELD_WIDTH/2.0 + partnery) < (FIELD_WIDTH - GOAL_WIDTH)/2.0) ) {
      prop.grade.downgrade();
    }
    // partner too far behind
    if (partnerx - playerx < -20) {
      prop.grade.downgrade();
    }    
    // pass should take the ball nearer to the goal
    if (sqrnorm(partner - otherGoal.pos) > sqrnorm(player - otherGoal.pos)) {
      prop.grade.downgrade();
    }
  }

  // if the player is the goalie the pass should not cross
  // the middle
  if(isGoalie)
    {
      if( playerx < 0 && partnerx > -FIELD_WIDTH/6)
	prop.grade.downgrade();
      if( playerx >= 0 && partnerx < FIELD_WIDTH / 6)
	prop.grade.downgrade();
    }

  return prop.grade;
}

////////////////////////////
//
// PassEval::betterPass
//
/////////////////////////// 

bool PassEval::betterPass(const Snapshot *wS, PropOf good, PropOf bad) const
{
  if (good.grade.is_better_than(bad.grade))
    return true;
  if (bad.grade.is_better_than(good.grade))
    return false;

  // ok, so the grades are the same
  if (good.grade.is_BAD())
    return false;

  int goodsum = 0;
  int badsum = 0;
  
  // player freestanding is better : 1 point
  if (good.free > bad.free)
    goodsum++;
  else if (bad.free > good.free)
    badsum++;

  bool defensive = ( min(good.pos.getX(), bad.pos.getX()) < 0.0 );


  // [C] und nochmal so ein paar Konstanten
  // broader corridor : 4/2 points
  if (good.corridor > bad.corridor)
    goodsum += ( defensive ? 4 : 2 );
  else if(bad.corridor > good.corridor)
    badsum += ( defensive ? 4 : 2 );

  // larger x-value of pass target : 2/4 points
  if( good.pos.getX() > bad.pos.getX() )
    goodsum += ( defensive ? 8 : 4 );
  else
    badsum += ( defensive ? 8 : 4 );
  
  // better goalkicker : 8 points
  if (good.number != bad.number) {
    Grade goodGoalGrade;
    Grade badGoalGrade;

    if ( goalevalpoints->isEmpty() )
      // no points were evaluated
      {
	goalevalpoints->setTime(wS->time);
	goodGoalGrade = (goalshoteval->ourEval(wS, good.pos));
	badGoalGrade  = (goalshoteval->ourEval(wS, bad.pos));
	goalevalpoints->insert(good.pos, goodGoalGrade);
	goalevalpoints->insert(bad.pos, badGoalGrade);
      }
    else
      {
	if (goalevalpoints->getTime() != wS->time)
	  // the evaluated points are from an earlier timestep
	  {
	    goalevalpoints->deleteAll();
	    goalevalpoints->setTime(wS->time);
	    goodGoalGrade = (goalshoteval->ourEval(wS, good.pos));
	    badGoalGrade  = (goalshoteval->ourEval(wS, bad.pos));
	    goalevalpoints->insert(good.pos, goodGoalGrade);
	    goalevalpoints->insert(bad.pos, badGoalGrade);
	  }
	else
	  {
	    if ( !goalevalpoints->allreadyComputed(good.pos, goodGoalGrade))
	      // the evaluation for good.pos was not done until now
	      {
		goodGoalGrade = (goalshoteval->ourEval(wS, good.pos));
		goalevalpoints->insert(good.pos, goodGoalGrade);
	      }
	    else 
	      {
		PAEVTRACE({ostrstream ost;
		ost << " betterPass: pos " << good.pos 
		    << " allready evaluated" << char(0);
		Comm::sendMCmd(MLog(ost.str()));
		} );
	      }
	    if( !goalevalpoints->allreadyComputed(bad.pos, badGoalGrade) )
	      // the evaluation for bad.pos was not done until now
	      {
		badGoalGrade = (goalshoteval->ourEval(wS, bad.pos));
		goalevalpoints->insert(bad.pos, badGoalGrade);
	      }
	    else
	      {
		PAEVTRACE( {
		  ostrstream ost;
		  ost << " betterPass: pos " << bad.pos 
		      << " allready evaluated" << char(0);
		  Comm::sendMCmd(MLog(ost.str()));
		} );
	      }
	  }
      }
    
    if (goodGoalGrade.is_better_than(badGoalGrade))
      goodsum += 8;
    else if (badGoalGrade.is_better_than(goodGoalGrade))
      badsum += 8;
  }

  if (goodsum > badsum)
    return true;
  return false;
}
