
//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

Grade PositionModule::goalieEvaluate(AdvSkills* skills){

  Grade g;

  // find up to 3 opponents that might shoot to goal

  priority = 1;
  if (wS->ball.pos.getX()>-FIELD_LENGTH/2+1.5*PENALTY_AREA_LENGTH) priority++;
  if (wS->ball.pos.getX()>-FIELD_LENGTH/2+2.5*PENALTY_AREA_LENGTH) priority++;
  if (wS->ball.pos.getX()>-FIELD_LENGTH/2+3.5*PENALTY_AREA_LENGTH) priority++;

  if (!me.inOurPenaltyArea())
    priority = 0;

  Vector2d potentialShootPosition[3]; 
  int potentialShooter[3];
  int reachBallTime[3];

  for (int i=0; i<3; i++){
    reachBallTime[i] = 10000;
    potentialShootPosition[i].setX(-FIELD_LENGTH/2+2*PENALTY_AREA_LENGTH);
    potentialShootPosition[i].setY(0);
    potentialShooter[i] = -1;
  }
  
  // look for opponents that might shoot direct
  int s=0;
  
  Vector2d guessedPos[12]; 
  for (int i=1; i<12; i++)
    guessedPos[i] = guessOppPos(i);

  if (tS->firstDangerousOpponent != UNKNOWN){
    if (guessedPos[tS->firstDangerousOpponent].getX() < -FIELD_LENGTH/2+2*PENALTY_AREA_LENGTH){  
      reachBallTime[0] = tS->otherTeamReachBallTime(tS->firstDangerousOpponent);
      potentialShootPosition[0] = putInsideField(tS->otherTeamReachBallPos(tS->firstDangerousOpponent));
      potentialShooter[0] = tS->firstDangerousOpponent; 
      s=1;
    }
  }
  
  for (int i=s; i<3; i++)
    for (int j=1; j<12; j++)
      if ( (j!=potentialShooter[0]) &&
	   (j!=potentialShooter[1]) && 
	   (j!=potentialShooter[2])){
	if (dist(guessedPos[j],middleOfGoal) < dist(potentialShootPosition[i],middleOfGoal)){
	  if (tS->firstDangerousOpponent != UNKNOWN)
	    reachBallTime[i] = tS->otherTeamReachBallTime(tS->firstDangerousOpponent) 
	      + minBallTime(dist(guessedPos[tS->firstDangerousOpponent] 
				 ,guessedPos[j] ));
	  else
	    reachBallTime[i] = tS->otherTeamReachBallTime(j);
	  if (reachBallTime[i]<25){
	    potentialShootPosition[i] = guessedPos[j]; 
	    potentialShooter[i] = j;
	  }
	}
      }

#ifdef GOALIETRACE
  for (int i=0; i<3; i++)
    if (potentialShooter[i] != -1)
      GOALIETRACE("pot. Shooter " << i << " : "<<potentialShooter[i] 
		  <<" ; reachBallTime: "<<reachBallTime[i]
		  <<" ; ShootPosition: "<<potentialShootPosition[i]);
#endif
  
#ifdef POSDRAWTRACE
  for (int i=0;i<3;i++)
    if (potentialShooter[i] != -1){
      POSDRAWTRACE(LINE(potentialShootPosition[i],
			Vector2d(-FIELD_LENGTH/2,-GOAL_WIDTH/2)));
      POSDRAWTRACE(LINE(potentialShootPosition[i],
			Vector2d(-FIELD_LENGTH/2,GOAL_WIDTH/2)));
    }
#endif
  
  if (potentialShooter[0] == -1){
    priority = 2;
    if (!me.inOurPenaltyArea())
      priority = 1;
    posGoal.setY(0);
    posGoal.setX(-FIELD_LENGTH/2 + 0.4*PENALTY_AREA_LENGTH);
    goToPrecision = CATCHABLE_AREA_W;  // why not?
    g.set_GOOD();
    GOALIETRACE("No Danger; want to go to standardPos " << posGoal);
    return g;
  }
  

  const int wGrid = (int) (2*(1.5*GOAL_WIDTH) / CATCHABLE_AREA_L); 
  const int lGrid = (int) (2*(0.5*PENALTY_AREA_LENGTH - CATCHABLE_AREA_L) 
			   / CATCHABLE_AREA_L); 
  double wStep = (1.5*GOAL_WIDTH) / wGrid;
  double lStep = (0.5*PENALTY_AREA_LENGTH - CATCHABLE_AREA_L) / lGrid;
  
  goToPrecision=min(lStep,wStep); //macht das Sinn???
  
  double maxMeetDist = -10*FIELD_LENGTH;
  double posCatchDist = 10*FIELD_LENGTH;
  
  Vector2d bestPos = potentialShootPosition[0] 
    + 0.5 * (middleOfGoal - potentialShootPosition[0]);
  
  Snapshot hSnap(*wS), *simSnap(&hSnap);

  for (int i=0; i<3; i++)
    if (potentialShooter[i]!=-1)
      posCatchDist = min(posCatchDist, 
			 minCatchDist(simSnap, oldPosGoal, 
				      potentialShootPosition[i]));

  for (int l=0; l<=lGrid; l++){
    double x = -FIELD_LENGTH/2 + l*lStep + CATCHABLE_AREA_L; 
    //if (x < potentialShootPosition[0].getX()){
    //  double a = (x - potentialShootPosition[0].getX()) / 
    //	(-FIELD_LENGTH/2 - potentialShootPosition[0].getX()); 
    //Vector2d left = potentialShootPosition[0] + 
    //a * (leftCorner - potentialShootPosition[0]);
    //Vector2d right = potentialShootPosition[0] + 
    //a * (rightCorner - potentialShootPosition[0]);
    for (int w=0; w<=wGrid; w++){
      double y = -0.75*GOAL_WIDTH + w*wStep;
      //if ((y>left.getY()) 
      //  && (y<right.getY())){
      Vector2d goaliePos=Vector2d(x,y);
      double posMinDist=10*FIELD_LENGTH;
      for (int i=0; i<3; i++){
	if ( (potentialShooter[i]!=-1) &&
	     (posMinDist>-FIELD_LENGTH/2))
	  posMinDist=min(posMinDist,
			 minCatchDist(simSnap, goaliePos, 
				      potentialShootPosition[i]));
      }
      if (posMinDist>maxMeetDist){
	maxMeetDist = posMinDist;
	bestPos = goaliePos;
      }
    }
    // }
    // }
  }

 
  if (maxMeetDist < -FIELD_LENGTH/2){
    bestPos = middleOfGoal + 
      (0.9 * CATCHABLE_AREA_L / (norm(potentialShootPosition[0] - middleOfGoal)))
      * (potentialShootPosition[0]-middleOfGoal);
    GOALIETRACE("maxMeetDist behind goal, bestPos Changed");
  }
  
  POSDRAWTRACE(POSITION(bestPos));
  
   
  double useableStam=calcUseableStamina();

  Angle strafAngle = PI/2;
  if (me.pos.getY() > wS->ball.pos.getY()) 
    strafAngle = 1.5 * PI;

  AdvSkills::TurtleCommand tc 
    = skills->goToPositionTimed(wS, oldPosGoal,useableStam, goToPrecision);
  // if backwards, double stamina
  if(tc.state == AdvSkills::TGOINGBACKWARDS)
    tc = skills->goToPositionTimed(wS, oldPosGoal,useableStam, goToPrecision);
			      
  int timeToReachOldPos = tc.time + 1;
  if (tc.state == AdvSkills::TFINISHED){
    AdvSkills::TurnCommand tuc 
      = skills->turn(wS,(wS->ball.pos - me.pos).getAngle()+strafAngle);
    if(tuc.state == AdvSkills::RFINISHED)
      timeToReachOldPos = 0;
    else
      timeToReachOldPos = 1;
  }
  

  tc = skills->goToPositionTimed(wS, bestPos,useableStam, goToPrecision);
  // if backwards, double stamina
  if(tc.state == AdvSkills::TGOINGBACKWARDS)
    tc = skills->goToPositionTimed(wS, bestPos,useableStam, goToPrecision);

  int timeToReachBestPos = tc.time + 1;
  if (tc.state == AdvSkills::TFINISHED){
    AdvSkills::TurnCommand tuc 
      = skills->turn(wS,(wS->ball.pos - me.pos).getAngle()+ strafAngle);
    if(tuc.state == AdvSkills::RFINISHED)
      timeToReachBestPos = 0;
    else
      timeToReachBestPos = 1;
  }  
  
 	       
  GOALIETRACE("time to best pos : "<< timeToReachBestPos 
	      <<" time to old pos : " << timeToReachOldPos);

  if ( (timeToReachBestPos < timeToReachOldPos + reachBallTime[0]) ||
       (posCatchDist< - FIELD_LENGTH/2)){

    posGoal = bestPos;
#ifdef GOALIETRACE
    if (timeToReachBestPos<timeToReachOldPos + reachBallTime[0])
      GOALIETRACE("Best Pos nearer OldPos + RBT[0]");
    if (posCatchDist<0)
      GOALIETRACE("posCatchDist<0");
    if ( maxMeetDist - posCatchDist > 2*PLAYER_SPEED_MAX )
      GOALIETRACE("maxMeetDist - posCatchDist > PLAYER_SPEED_MAX");
#endif
    GOALIETRACE("Changed go to Position: " << posGoal 
		<< "would get Ball at" << maxMeetDist);
    POSDRAWTRACE(LINE(Vector2d(maxMeetDist,-PENALTY_AREA_WIDTH/2),
		      Vector2d(maxMeetDist,PENALTY_AREA_WIDTH/2)));
  } else {
    
    posGoal = oldPosGoal;
   
    GOALIETRACE("Staying cool, still going to " << posGoal 
		<< "would get Ball at" << posCatchDist);
    POSDRAWTRACE(LINE(Vector2d(posCatchDist,-PENALTY_AREA_WIDTH/2),
		      Vector2d(posCatchDist,PENALTY_AREA_WIDTH/2)));  
  }
 
  g.set_GOOD(); 
  
  return g;
} 


bool PositionModule::GoalieAct(AdvSkills* skills)
{
  GOALIETRACE("GoalieAct: going to " << posGoal);
  double useableStam=calcUseableStamina();
  
  AdvSkills::TurtleCommand tc 
    = skills->goToPosition(wS, posGoal, goToPrecision,useableStam,
			   true, (wS->ball.pos - me.pos).getAngle());
  // if backwards, double stamina
  if(tc.state == AdvSkills::TGOINGBACKWARDS)
    tc = skills->goToPosition(wS, posGoal, goToPrecision, 
			      min(2*useableStam,availableStamina),
			      true, (wS->ball.pos - me.pos).getAngle());
  
  Angle strafAngle = PI/2;
  if (me.pos.getY() > wS->ball.pos.getY()) 
    strafAngle =  1.5 * PI;
  GOALIETRACE("strafAngle : " << strafAngle);
  Angle turnAngle=(wS->ball.pos - me.pos).getAngle();
  if (me.pos.getX()<-FIELD_LENGTH/2 + 1.5 * CATCHABLE_AREA_L)
    turnAngle = 0;
  if (tc.state == AdvSkills::TFINISHED){// I am at my desired position => turn to ball
    AdvSkills::TurnCommand tuc 
      = skills->turn(wS,turnAngle+strafAngle);
    //desiredHeadAng = (wS->ball.pos - me.pos).getAngle();
    
    if(tuc.state == AdvSkills::RFINISHED)
      skills->noAction();
    else
      skills->exec(wS, tuc);
  }
  else{ // skills not finished
    skills->exec(wS,tc);
    
  }

  return true;
}

int PositionModule::minBallTime(double distance){
  int i=0;
  while (distance>ballDist[i] && i<50)
    i++;
  return i;
}


double PositionModule::minCatchDist(Snapshot* simSnap, Vector2d goaliePos, Vector2d shootPos){
  WorldBall rMeet,lMeet;
  simSnap->my.pos = goaliePos;
  simSnap->my.bodyDir = (wS->ball.pos - simSnap->my.pos).getAngle();
  simSnap->my.v = Vector2d(0,0);
  simSnap->ball.pos = shootPos;
  simSnap->ball.v = (BALL_SPEED_MAX / norm(rightCorner - shootPos) ) 
    * (rightCorner - shootPos);
  simSnap->ball.simulate(); //need some time to see the ball
  AdvSkills::TurtleCommand tc = 
    skills->controlBall(simSnap,
			wS->senseBody.stamina-STAMINA_MAX*RECOVER_DEC_THR,
			0.9*CATCHABLE_AREA_L,rMeet);
  simSnap->my.bodyDir = (wS->ball.pos - simSnap->my.pos).getAngle();
  simSnap->ball.pos = shootPos;
  simSnap->ball.v = (BALL_SPEED_MAX / norm(leftCorner - shootPos) ) 
    * (leftCorner - shootPos);
  simSnap->ball.simulate(); //need some time to see the ball
  tc = skills->controlBall(simSnap,
			   wS->senseBody.stamina-STAMINA_MAX*RECOVER_DEC_THR,
			   0.9*CATCHABLE_AREA_L,lMeet);
  return min(rMeet.pos.getX(),lMeet.pos.getX());
}


Vector2d PositionModule::guessOppPos(int nr){
  
 return wS->otherTeam(nr).pos; 

//   int age = min(5,wS->otherTeam(nr).age);
//    Vector2d guessedPos = wS->otherTeam(nr).pos;
//    Vector2d target=guessedPos;

 

//    if (guessedPos.getX() < -FIELD_LENGTH/2 + 0.6*PENALTY_AREA_LENGTH) // we dont suspect opps to com close
//      if (guessedPos.getY() < -PENALTY_AREA_WIDTH/2-2)
//        target = Vector2d(guessedPos.getX(),-PENALTY_AREA_WIDTH/2-2);
//      else 
//        if (guessedPos.getY() > PENALTY_AREA_WIDTH/2-2)
//        target = Vector2d(guessedPos.getX(),PENALTY_AREA_WIDTH/2-2);
//        else
//  	return guessedPos;  // player might stand still
//    else 
//      if (guessedPos.getY() < -PENALTY_AREA_WIDTH/2-2)
//        target = Vector2d(-FIELD_LENGTH/2 + 0.6 * PENALTY_AREA_LENGTH,-PENALTY_AREA_WIDTH/2-2);
//      else 
//        if (guessedPos.getY() > PENALTY_AREA_WIDTH/2-2)
//  	target = Vector2d(-FIELD_LENGTH/2 + 0.6 * PENALTY_AREA_LENGTH,PENALTY_AREA_WIDTH/2-2);
//        else
//  	target = Vector2d(-FIELD_LENGTH/2 + 0.6 * PENALTY_AREA_LENGTH,guessedPos.getY()); 

//    while ( (age>0) && (dist(guessedPos,target) > wS->otherTeam(nr).vmax )){ 
//      guessedPos += (wS->otherTeam(nr).vmax / norm (target - guessedPos)) * 
//        (target - guessedPos);
//      age--;  
//    }
//    return guessedPos;
}

Vector2d PositionModule::putInsideField(Vector2d pos){
  Vector2d newPos=pos;
  if (pos.getX()<-FIELD_LENGTH/2+1){
    newPos *= (-FIELD_LENGTH/2+1) / pos.getX();
    GOALIETRACE("Pos outside; new Pos:"<< newPos);
  }
  if (pos.getX()>FIELD_LENGTH/2-1){
    newPos *= (FIELD_LENGTH/2-1) / pos.getX();
    GOALIETRACE("Pos outside; new Pos:"<< newPos);
  } 
  if (pos.getY()<-FIELD_WIDTH/2+1){
    newPos *= (-FIELD_LENGTH/2+1) / pos.getY(); 
    GOALIETRACE("Pos outside; new Pos:"<< newPos);
  }
  if (pos.getY()>FIELD_WIDTH/2-1){
    newPos *= (FIELD_LENGTH/2-1) / pos.getY();
    GOALIETRACE("Pos outside; new Pos:"<< newPos);
  }
 
  return newPos;

}
