//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module PositionModule
//
// $Id: <add a dollar character here>
// $manager:
//
//////////////////////////////////////////////////////////////////////


#ifndef __POSITIONMODULE_H__
#define __POSITIONMODULE_H__

#define FUNSSELMACROS

#include "grade.h"
#include "PositionEval.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "goalshoteval.h"
#include "passeval.h"
#include "drawMacros.h"
#include "Config.h"

//um Raumdeckung zu aktivieren, diese Zeile auskommentieren
#define Raumdeckung

//der alte Goalie
//#define AlterTorwart

// Trace macro: insert your own short name for your module
// uncomment the line you want.
#undef POSTRACE
//#define POSTRACE(output) cout << wS->time << " PositionModule: " << output << endl

#undef POSTRACESHORT
// use this line if you want prompting of players decision process
//#define POSTRACESHORT(output) cout << " POS: " << output << endl; 

#undef POSDRAWTRACE
// use this if you want some modern art in the replayer 
//#define POSDRAWTRACE(output) Comm::sendMCmd(output)

#undef GOALIETRACE 
//#define GOALIETRACE(output) { ostrstream ost; ost << "GOALIETRACE: " << output; Comm::sendMCmd(MLog(ost.str()));} 

#ifndef POSTRACE
#define POSTRACE(output)
#else
#warning POSTRACE enabled
#endif
#ifndef POSTRACESHORT
#define POSTRACESHORT(output)
#else
#warning POSTRACESHORT enabled
#endif
#ifndef POSDRAWTRACE
#define POSDRAWTRACE(output)
#else
#warning POSDRAWTRACE enabled
#endif
#ifndef GOALIETRACE
#define GOALIETRACE(output)
#else
#warning GOALIETRACE enabled
#endif

//////////////////////////////////////////////////////////////////////
//
// definitions of constants
//
//////////////////////////////////////////////////////////////////////

 const double COVER_DISTANCE = 1.5*VISIBLE_DISTANCE; // in this distance an other player is covered
 const int NumberofPointsX = 2;
 const int NumberofPointsY = 5;
//number of points checked when searching for free position
//////////////////////////////////////////////////////////////////////
//
// enum PosBallControlState
//
//////////////////////////////////////////////////////////////////////

//what actions are known
enum plannedAction {
  DEFAULT,
  NO_ACTION,           //planning to let an other module have control 
  //over the player
  SUPPORT_RESP_PLAYERS, // 
  PLAYER_ATTACK,       //attack opponent with the ball
  GOTO_FREE_POSITION,  
  COVER_OPPONENT
};


//Forward declaration
class PositionModuleInfo;


class PositionModule
{
public:
  /** constructor of module PositionModule */
  PositionModule(bool isGoalie, const PassEval *_passEval, 
		 const GoalShotEval *_goalShotEval, 
		  PositionEval *_positionEval);

  ~PositionModule();
  
  /** feed information about new world and tactics into module.
   */
  bool feedbackWorld(const Snapshot* _wS, TacticalSnapshot* _tacticalSnap);

  /** Feed audio information into module. This method is called for
   * each audio message. audioMsg.getSource() is player No or 0 for coach
   * Audiomsg audioInfo is acting like a stream.
   * usage: audioMsg >> auInfo (where PositionModuleInfo auInfo holds the information
   * for module PositionModule in other players)
   */
  bool feedbackAudio(AudioMsg &audioMsg);

  /** Do the evaluation for the next action
   * For definition of Grade: see grade.h
   */
  Grade evaluate(AdvSkills* skills,
                 const Snapshot* _wS, TacticalSnapshot* _tS);

  /** Decide whether a turn head should be done.
   * This is always called after act, so the data of act and
   * evaluate should be used to calculate the TurnHeadCommand
   * which is then executed by the Mastercontrol.
   * skills should only be used to calculate the command,
   * NOT TO SEND IT!!
   */
  Grade turnHeadEvaluate(AdvSkills* skills, AdvSkills::TurnHeadCommand &thc);
  
  /** Perform the action for this module */
  bool act(AdvSkills* skills,
           const Snapshot* _wS, TacticalSnapshot* _tS);

  /** description of stream AudioMsg: see above. Here you should implement:
   * audioMsg << PositionModuleInfo;
   */
  bool getSayMsg(AudioMsg &audioMsg);

private:
  bool isGoalie; //am I goalie?

  const Snapshot* wS;  //where the snapshots are saved
  TacticalSnapshot* tS;

  // Where do I want to go today?!
  Vector2d posGoal;
  double goToPrecision;//how fast do I want to go there?
  int priority;
  // save old posGoal
  Vector2d oldPosGoal;
  // middle of our goal (just for the goalie)
  Vector2d middleOfGoal;
  Vector2d leftCorner, rightCorner;
  // to store the messages that shall be send
  PositionModuleInfo *sendInfo;

  // own offside lines
  double ourOffsideLine;
  double otherOffsideLine;

  WorldPlayer me;
  int nearestFriendNumber;
  int internalTime;
  int myTactNo;
  int nearestOther;

  // the opponent we cover (to look at him from time to time)
  // -1 means no opponent to cover
  //int coveredOpponent; 
  const WorldMovedObject * lookObject;
  Vector2d lookPos;

 // the agents modules 
  const PassEval     *passEval;
  const GoalShotEval *goalShotEval;
  PositionEval *positionEval;

  //what do I want to do
  plannedAction action;

  // I need to know what I did for TurnHeadEvaluate
  bool didIAct;
  Grade EvalGrade;

 
  bool searchFreePosition();
  //decides whether the player should look for a free position or not

  bool evalSupportRespPlayer();
  //decides whether the player should support the resp. players
  int SupportPlayerNo;
  //which player should be supported
  void SupportPlayer();


  double availableStamina; //how much stamina may max. be used 
  double calcUseableStamina();
  //is used when we have to save stamina
  void CalcStdPriority();
  //calculates the priority with which the player should run to it's 
  //position


  //////////////////////////////////////////////////////////////////////
  //                     The Goalie                                   //
  /////////////////////////////////////////////////////////////////////
  Grade goalieEvaluate(AdvSkills* skills);

  bool GoalieAct(AdvSkills* skills);

  double minCatchDist(Snapshot* simSnap, Vector2d goaliePos, Vector2d shootPos);

  int minBallTime(double distance);  // to determine the time needs for distance
  // with maximum speed at the beginning
  double ballDist[50]; // needed to determine minBallTime

  Vector2d guessOppPos(int nr);

  Vector2d putInsideField(Vector2d pos); // if pos outside Field
  // we intersect pos with field borders
  //////////////////////////////////////////////////////////////////


  // to attack opponents with the ball
  Vector2d goAroundOpponent(Vector2d target,WorldPlayer opponent,
			    double dist, int inside);

  void stdact();

  void playerAttack(int Opponent);
  int taReportTime;
  
  double GradeToDouble(Grade value);
  Grade DoubleToGrade(double value);
  //converts grade into a number and a number into a grade
  Vector2d gotoFreePos();
  Vector2d LastBestPos;
    //looks for a good pass position 
  int lastActive;
  //time when FreePos was used the last time 
  Vector2d middlePositions[NumberofPointsX*NumberofPointsY+1] ;
  double sizeOfMiddleGrades[NumberofPointsX*NumberofPointsY+1];
  // while searching for a free position we check best pass positions in the 
  // playerarea, now take the middle of these position for several steps 
  
  void coverOpponent();
  Vector2d CalCoverPos(int CoveredOpp);

  // Is this player free ?
  bool otherPlayerFree(const Snapshot *wS,int num, 
		       double distance = COVER_DISTANCE);
  bool ourPlayerFree(const Snapshot *wS,int num, 
		     double distance = COVER_DISTANCE);

  void respectStandardSituations(const Snapshot* wS,
				 const TacticalSnapshot* tS);
  // not very sophisticate method to avoid collisions with other players
  void avoidPlayerCollision(const Snapshot *wS);

  Vector2d calculateCoverPosFromPlayerNo(const Snapshot *wS, int plNo);
}; // class PositionModule

//////////////////////////////////////////////////////////////////////
// Class to keep information for modules of other agents.
//////////////////////////////////////////////////////////////////////

class PositionModuleInfo
{
public:

  /** constructor of Info class for module PositionModule.
   * Here you should implement 
   * valid = _valid;
   */
  PositionModuleInfo(bool _valid);
  ~PositionModuleInfo();

  /** flag to determine if instance of PositionModuleInfo is valid. */
  bool valid;
}; // class PositionModuleInfo

/** put information to AudioMsg-Stream.
 * Here you should implement
 * aumsg << info.valid;
 * if (info.valid) aumsg << // further data structures to write ;
 * return aumsg;
 */
class AudioMsg &operator<< (AudioMsg &aumsg, const PositionModuleInfo &info);

/** fill with information from AudioMsg-Stream.
 * Here you should implement
 * aumsg >> info.valid;
 * if (info.valid) aumsg >> // further data structures to read ;
 * return aumsg;
 */
class AudioMsg &operator>> (AudioMsg &aumsg, PositionModuleInfo &info);

#endif
