//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// \$Id:
// \$manager: gauer$
//
//////////////////////////////////////////////////////////////////////

#include "standardsituationmodule.h"
#include "Config.h"
#include "minmax.h"

int
StandardSituationModule::
calcClosestDesiredPlayersTo(const Vector2d target[], 
			    const int count,
			    const Snapshot *worldSnap,
			    const TacticalSnapshot *ts,
			    int players[]) const
{
  int off = 0;
  int mid = 0;
  int def = 0;
  

  int defender[10], d = 0;
  int midfield[10], m = 0;
  int offender[10], o = 0;
  int team[11];
  for (int i = 0; i < 10; i++)
    {
      defender[i] = UNKNOWN;
      midfield[i] = UNKNOWN;
      offender[i] = UNKNOWN;
    }

  // calc our defenders, midfielders and offenders
  for (int player = 1; player < 12; player++)
    {
      int tacNo = ts->ourTeamTacticalNo(player);
      if (tacNo != UNKNOWN)
	{
	  if (ts->ourTeamDefender(tacNo))
	    defender[d++] = player;
	  if (ts->ourTeamMidfield(tacNo))
	    midfield[m++] = player;
	  if (ts->ourTeamOffender(tacNo))
	    offender[o++] = player;
	}
    }  

  OTHERTRACE( " known defenders: " << d );
  CMDTRACE( ost << "       "; for (int i = 0; i < d; i++) { ost << defender[i] << " "; } );
  OTHERTRACE( " known midfield: " << m );
  CMDTRACE( ost << "       "; for (int i = 0; i < m; i++) { ost << midfield[i] << " "; } );
  OTHERTRACE( " known offenders: " << o );
  CMDTRACE( ost << "       "; for (int i = 0; i < o; i++) { ost << offender[i] << " "; } );

  calcDesiredPlayerTypes(worldSnap, ts, worldSnap->ball.pos, def, mid, off);
  
  // if we want more defenders than we know
  if (def-d > 0)
    {
      DECISIONTRACE( " not enough defenders known, taking midfielders " );
      def = d;
      mid += def-d;
    }
  
  // if we want more offenders than we know
  if (off-o > 0)
    {
      DECISIONTRACE( " not enough offenders known, taking midfielders " );
      off = 0;
      mid += off-o;
    }
  
  // if we want more midfielders than we know
  if (mid-m > 0)
    {
      DECISIONTRACE( " not enough middfielders known " );
      mid = m;
      if (d > def && worldSnap->ball.pos.getX() <= 0)
	{
	  DECISIONTRACE( " taking a defender " );
	  def++;
	}
      else if (o > off && worldSnap->ball.pos.getX() > 0)
	{
	  DECISIONTRACE( " taking an offender " );
	  off++;
	}
    }

  // calc the possible players
  int n = 0;
  for (int i = 0; i < ((def > 0) ? d : 0); i++)
    team[n++] = defender[i];
  for (int i = 0; i < ((mid > 0) ? m : 0); i++)
    team[n++] = midfield[i];
  for (int i = 0; i < ((off > 0) ? o : 0); i++)
    team[n++] = offender[i];

  int actPlayers(min(count, def+mid+off));
  for (int i = 0; i < actPlayers; i++) {
    CMDTRACE(ost << "looking for players in: "; 
	     for (int i = 0; i < n; i++)
               { ost << team[i] << " "; }
            );
    
    // calc the closest player for each target
    calcClosestPlayerTo(target[i], team, n, worldSnap, players[i]);

    int tacNo(ts->ourTeamTacticalNo(players[i]));
    if (ts->ourTeamDefender(tacNo)) def--;
    if (ts->ourTeamMidfield(tacNo)) mid--;
    if (ts->ourTeamOffender(tacNo)) off--;
    
    // calc the possible players
    n = 0;
    for (int j = 0; j < ((def > 0) ? d : 0); j++) {
      bool used(false);
      for (int k = 0; k <= i; k++)
	if (defender[j] == players[k]) {
	  used = true;
	  break;
	}
      if (!used) team[n++] = defender[j];
    }
    for (int j = 0; j < ((mid > 0) ? m : 0); j++) {
      bool used(false);
      for (int k = 0; k <= i; k++)
	if (midfield[j] == players[k]) {
	  used = true;
	  break;
	}
      if (!used) team[n++] = midfield[j];
    }
    for (int j = 0; j < ((off > 0) ? o : 0); j++) {
      bool used(false);
      for (int k = 0; k <= i; k++)
	if (offender[j] == players[k]) {
	  used = true;
	  break;
	}
      if (!used) team[n++] = offender[j];
    }
  }

  calcShortWays(worldSnap, players, target, actPlayers);

#ifdef USEDRAWTRACE
  for (int i = 0; i < actPlayers; i++) {
    ostrstream text;
    text << players[i] << (char)0;
    DRAWTRACE( TEXT(text.str(), target[i]+Vector2d(.5,0)) );
  }
#endif

  return actPlayers;
}

void
StandardSituationModule::
calcDesiredPlayerTypes(const Snapshot *worldSnap, 
		       const TacticalSnapshot *ts,
		       const Vector2d &target, int &noDef, 
		       int &noMid, int &noOff) const
{
  // calc the indices to access the array with the number of player types
  // for the numbers see readAgent_conf()
  int team = 0;
  int fieldPart = 0;
  int tacticalSystem = 0;

  switch (worldSnap->playMode)
    {
      // we have the ball
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_OFFSIDE_THEM:
      team = 0;
      break;
    // they have the ball
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_KICK_IN_THEM:
    case PLAY_MODE_CORNER_KICK_THEM:
    case PLAY_MODE_OFFSIDE_US:
      team = 1;
      break;
    default: ;
    }

  switch (getOurTacticalSystem(ts))
    {
    case 442:
      tacticalSystem = 0;
      break;
    case 352:
      tacticalSystem = 1;
      break;
    case 343:
      tacticalSystem = 2;
      break;
    default: ;
    }
  
  double maxX(positionEval->GetMaxX()), minX(positionEval->GetMinX());
  double d(fabs(maxX-minX)/3.);

  DRAWTRACE( VLINE(maxX) );
  DRAWTRACE( VLINE(minX) );

  // [C]*
  // if the target is in left third between minX and maxX
  if (target.getX() <= minX+d+5.)
    fieldPart = 0;

  // if the target is in the middle third of the field
  if (target.getX() <= minX+2.*d+5. &&
      target.getX() > minX+d)
    fieldPart = 1;

  // if the target is in their third of the field
  if (target.getX() >= minX+2.*d)
    fieldPart = 2;

  noDef = players[team][fieldPart][tacticalSystem].defenders;
  noMid = players[team][fieldPart][tacticalSystem].midfield;
  noOff = players[team][fieldPart][tacticalSystem].offenders;

  DECISIONTRACE( " Desired player types: defenders: " << noDef );
  DECISIONTRACE( "                        midfield: " << noMid );
  DECISIONTRACE( "                       offenders: " << noOff );
  
}

void
StandardSituationModule::
calcClosestPlayerTo(const Vector2d &target, 
		    const int team[], const int count, 
		    const Snapshot *ws, 
		    int &player) const
{
  // calc distances
  double dists[11];
  
  for (int i = 0; i < 11; i++)
    dists[i] = NOWHERE_DIST; 

  for (int i = 0; i < count; i++)
    {
      if (team[i] == UNKNOWN)
	continue;
      dists[team[i]-1] = sqrnorm(ws->ourTeam(team[i]).pos-target);
    }
  // make sure the goalie is not the closest player
  if (ws->ourGoalieNo != UNKNOWN)
    dists[ws->ourGoalieNo-1] = 2 * NOWHERE_DIST; 
  
  // find closest player
  player = UNKNOWN;
  double dist = 2 * NOWHERE_DIST; 
  for (int i = 0; i < 11; i++)
    {
      if (dists[i] < dist)
	{
	  dist = dists[i];
	  player = i+1;
	}
    }

  // find players close to target too
  dist = sqr(sqrt(dist)+stabilitySearchDist); 
  int closePlayers[11], closeCount = 0;
  
  for (int i = 0; i < 11; i++)
    {
      if (dists[i] < dist)
	closePlayers[closeCount++] = i+1;
    }

  // the closest player is the close player with 
  // smallest player number or the closest
  if (closeCount > 0)
    player = closePlayers[0];
}

Vector2d
StandardSituationModule::
calcPossibleGoalPoint(const Snapshot *worldSnap) const
{
  // The left and the right edges of our goal
  Vector2d leftPost(worldSnap->ourGoal.pos.getX(),
		    worldSnap->ourGoal.pos.getY()+GOAL_WIDTH/2),
    rightPost(worldSnap->ourGoal.pos.getX(),
	      worldSnap->ourGoal.pos.getY()-GOAL_WIDTH/2);

  // calc the intersection of the bisector of the angle between the lines 
  // from the ball to the edges of the goal and
  // the line between the edges of the goal
  Vector2d intersect(rightPost-worldSnap->ball.pos);

  double a = sqrnorm(leftPost-worldSnap->ball.pos), 
    b = sqrnorm(rightPost-worldSnap->ball.pos),
    c = sqrnorm(leftPost-rightPost),
    alpha = acos((a+b-c)/(2.*sqrt(a*b)));
  
  intersect.turn(-alpha/2);
  
  double detA = intersect.getX()*(rightPost.getY()-leftPost.getY())+
    intersect.getY()*(rightPost.getX()-leftPost.getX());
  if (detA)
    {
      double lambda = 1/detA*((leftPost.getY()-rightPost.getY())*
			      (worldSnap->ball.pos.getX()-rightPost.getX())+
			      (rightPost.getX()-leftPost.getX())*
			      (worldSnap->ball.pos.getY()-rightPost.getY()));
      intersect.stretch(lambda);
    }
  
  intersect = intersect + worldSnap->ball.pos;  // point between the goal posts
  
  
  // if detA == 0 then the ball is on the line constructed by the goal posts
  // and a good place for the goalie is the goal post closest to the ball
  return intersect;
}

void
StandardSituationModule::
calcWayTo(const Vector2d &from, Vector2d &to, 
	  const double stamina,
	  const Vector2d &m, const double &r, 
	  const Snapshot *worldSnap) const
{
  if (worldSnap); // to avoid warnings, when no traces are active

  DRAWTRACE( CIRCLE(m, r) );
  OTHERTRACE( "Dist to target: " << norm(from-to) );
  // can we walk from "from" to "to" without entering the circle defined by m and r

  // if we are close to the target do nothing
  // [C]
  if (sqrnorm(to-from) <= sqr(2*PLAYER_SIZE)) 
    {
      to = from;
      return;
    }

//    // if I am inside the circle
//    double nrm = sqrnorm(from-m);
//    // [C]
//    if (nrm < sqr(r-PLAYER_SIZE))
//      {
//        DECISIONTRACE( " walking out of the circle" );
//        // if I am close to the middle of the circle
//        // [C]
//        if (nrm < sqr(1e-4))
//  	{
//  	  // go towards target
//  	  to -= from;
//  	  // [C]
//  	  to *= (worldSnap->my.dashPowerRate*stamina+norm(worldSnap->my.v)+1)/norm(to);
//  	  to += worldSnap->my.pos;
//  	}
//        else
//  	{ // take the shortest way out of the circle
//  	  to = from - m;
//  	  // [C]
//  	  to *= (DASH_POWER_RATE*stamina+norm(worldSnap->my.v)+1)/norm(to);
//  	  to += from;
//  	}
//        return;
//      }

  // [C]
  bool weAreClose = sqrnorm((from-m).stretch(r/norm(from-m))+m-from) 
    <= sqr(PLAYER_SIZE+0.1);
  Vector2d c = (from-m).stretch(r/norm(from-m));
  // if the direct way leeds through the circle
  // [C]
  if ((fabs(asin(r/norm(m-from))) -  
      fabs(angleBetween(m-from, (1-PLAYER_SIZE/norm(to-m))*(to-m)+m-from)) >=
      -1e-4) && (sqrnorm(to-from) > sqrnorm(m-from)))
    {
      DECISIONTRACE( " Can not walk directly to " << to );
      if (weAreClose)
	{
	  Vector2d rightPos = c;
	  rightPos.perp();
	  Vector2d leftPos = rightPos;
	  leftPos.stretch(-1.0);
	  rightPos += from;
	  leftPos += from;
	  // go tangential to the circle closer to the target
	  if ( sqrnorm(leftPos - to) < sqrnorm(rightPos - to ) )
	    {
	      to = leftPos;
	      DECISIONTRACE( "close to the ball, going to the left side" );
	    }
	  else
	    {
	      to = rightPos;
	      DECISIONTRACE( "close to the ball, going to the right side" );
	    }
	}
      else
	{
	  // we are not close to the ball and have to decide if we go left or right
	  Vector2d rightPos = c;
	  // [C]
	  rightPos.stretch(1.1);
	  rightPos.perp();
	  Vector2d leftPos = rightPos;
	  leftPos.stretch(-1.0);
	  leftPos += m;
	  rightPos += m;
	  OTHERTRACE( " leftPos: " << leftPos << " rightPos: " << rightPos);
	  OTHERTRACE( " c: " << c);

	  if ( sqrnorm(leftPos - to) < sqrnorm(rightPos - to) )
	    {
	      // go to the left pos
	      to = c;
	      to.turn( -fabs(acos(r / norm(m-from))) );
	      DECISIONTRACE( "not close to the ball, going to the left side" );
	    }
	  else
	    {
	      // go to the right pos
	      to = c;
	      to.turn( fabs(acos(r / norm(m-from))) ); 
	      DECISIONTRACE( " not close to the ball, going to the right side " );
	    }
	  // [C]
	  to.stretch( 1.1 );
	  to += m;
	  // stretch "to" to the length we can go + a little more for security
	  // goingTo() should not turn to ball before reaching the real target
	  to -= from;
	  // [C]
	  to *= (DASH_POWER_RATE*stamina+norm(worldSnap->my.v)+1)/norm(to);
	  to += from;
	  OTHERTRACE( " turnAngle: " << RAD_TO_DEG(fabs(acos(r / norm(m-from)))) 
		      << " to: " << to);
	}
    }
}

void
StandardSituationModule::
calcShortWays(const Snapshot* worldSnap,
	      int players[],
	      const Vector2d target[],
	      const int count) const
{
  if (count <= 1 || count > 5) return;

  Permutations P(Perm->of(count));
  Grid<double> dist(count, count);
  double d(0);

  // generate distance Matrix
  for (int p = 0; p < count; p++)
    for (int t = 0; t < count; t++) {
      d = norm(worldSnap->ourTeam(players[p]).pos-target[t]);
      dist.set(p,t,&d);
    }

  double D(NOWHERE_DIST);
  unsigned int perm(0);
  for (unsigned int i = 0; i < P.numPerm(); i++) {
    d = 0;
    for (int p = 0; p < count; p++) {
      d += *dist.get(p, P[i][p]);
      OTHERTRACE("perm " << i << " " << (int)P[i][p] << " dist: " << *dist.get(p, P[i][p]));
    }
    OTHERTRACE("Dist: " << d);
    if (D-2 > d) { //[C]
      D = d;
      perm = i;
    }
  }
  
  OTHERTRACE("chosen perm: " << perm << " with dist " << D);

  int Player[11];
  for (int i = 0; i < count; i++)
    Player[i] = players[P[perm][i]];
  for (int i = 0; i < count; i++)
    players[i]=Player[i];
}

Vector2d
StandardSituationModule::
calcIntersection(const Vector2d &A, const Vector2d &B, 
		 const Vector2d &C, const Vector2d &D) const
  // intersect the lines AB and CD ?
  // Rem.: AB not parallel to CD
{  

  Vector2d E(A);
  E -= C;
  Vector2d F(A);
  F -= B;
  Vector2d G(D);
  G -= C;

  OTHERTRACE( "Intersection: A=" << A << ", B=" << B << ", C=" << C << ", D=" << D )
  OTHERTRACE( "  E=" << E << ", F=" << F << ", G=" << G )
  

  double detM = F.getX()*G.getY() - F.getY()*G.getX();

  if (fabs(detM) < eps)
    {
      ERRORTRACE1( "det(M)= " << detM );
      return A;
    }

  double t = E.getX()*G.getY() - G.getX()*E.getY();
  double s = F.getX()*E.getY() - F.getY()*E.getX();

  t /= detM;
  s /= detM;

  Vector2d S1 = F;
  S1.stretch(1.0-t);
  S1 += B;

  Vector2d S2 = G;
  S2.stretch(s);
  S2 += C;

  // [C]
  if (sqrnorm(S1-S2) > sqr(0.2))
    ERRORTRACE1( "Error, norm = " << norm(S1-S2) );
  
  OTHERTRACE( "  t=" << t << ", s=" << s << ", S1=" << S1 << ", S2=" << S2 );

  return S1;
}

double
StandardSituationModule::
calcDistAroundCircle(const Vector2d &A, const Vector2d &B,
		     const Vector2d &M, const double r, 
		     const Snapshot *worldSnap) const
{
  if (worldSnap);        // to avoid warnings

  Vector2d MMinusA = M;
  MMinusA -= A;
  Vector2d MMinusB = M;
  MMinusB -= B;

  double alpha = asin(r/norm(MMinusA));
  double beta  = asin(r/norm(MMinusB));

  Vector2d leftPos  = MMinusA;
  Vector2d rightPos = MMinusA;
  leftPos.turn(alpha);
  leftPos += A;
  rightPos.turn(-alpha);
  rightPos += A;
  
  Vector2d hatA, hatB; 
  if (sqrnorm(leftPos - B) < sqrnorm(rightPos - B))
    {
      hatA = leftPos;
      hatB = MMinusB;
      hatB.turn(-beta);
    }
  else
    {
      hatA = rightPos;
      hatB = MMinusB;
      hatB.turn(beta);
    }

  hatA += A;
  hatB += B;

  Vector2d S = calcIntersection(A, hatA, B, hatB);


  DRAWTRACE( POSITION(leftPos) );
  DRAWTRACE( POSITION(rightPos) );

  DRAWTRACE( CIRCLE(M, r) );
  DRAWTRACE( LINE(A, S) );
  DRAWTRACE( LINE(B, S) );
  DRAWTRACE( POSITION(hatA) );
  DRAWTRACE( POSITION(hatB) );
  DRAWTRACE( POSITION(M) );
  
  OTHERTRACE( "calcDistAroundCircle: A=" << A << ", B=" << B
	      << ", hatA=" << hatA << ", hatB=" << hatB );

  return (norm(S-A) + norm(S-B));
}


int
StandardSituationModule::
calcOurKicksThemPositions(const Snapshot *worldSnap,
			  const TacticalSnapshot *,
			  Vector2d positions[]      ) const
{
  int players(0); // number of players needed to block every pass better than BAD
  Snapshot WS(*worldSnap), *myWS(&WS); // own copy of Snapshot
  WorldPlayer oplayer; // the player who is possibly going to kick

  DRAWTRACE( VLINE(myWS->otherOffsideLine(maxAge)) );

  // find the closest player to the ball of the other team 
  double dist(NOWHERE_DIST);
  int num(UNKNOWN);
  for (int i = 1; i < 12; i++)
    if (sqrnorm(myWS->otherTeam(i).pos-myWS->ball.pos) < dist) {
      dist = sqrnorm(myWS->otherTeam(i).pos-myWS->ball.pos);
      num = i;
    }
  oplayer = myWS->otherTeam(num);

  // set him above, below or behind the ball
  //[C]*
  if (myWS->ball.pos.getY() > FIELD_WIDTH/6.) {
    oplayer.pos = myWS->ball.pos + Vector2d(0,BALL_SIZE+PLAYER_SIZE);
    oplayer.bodyDir = Angle(-PI_2);
  }
  else if (myWS->ball.pos.getY() < -FIELD_WIDTH/6.) {
    oplayer.pos = myWS->ball.pos - Vector2d(0,BALL_SIZE+PLAYER_SIZE);
    oplayer.bodyDir = Angle(-3./2.*PI);
  }
  else {
    oplayer.pos = myWS->ball.pos + Vector2d(BALL_SIZE+PLAYER_SIZE,0);
    oplayer.bodyDir = Angle(-PI);
  }
  oplayer.faceDir = oplayer.bodyDir;
  myWS->setOtherTeam(num, oplayer);

#ifdef USEDRAWTRACE
  DRAWTRACE( POSITION(oplayer.pos) );
  char text[3]; sprintf(text, "%d", num);
  DRAWTRACE( TEXT(text, oplayer.pos+Vector2d(.5,0)) );
#endif

  //remove our players
  oplayer.pos = Vector2d(0, -FIELD_LENGTH/2-10);
  for (int i = 1; i < 12; i++) 
    myWS->setOurTeam(i, oplayer);

  // check if he can do a goalshot
  Grade goalGrade(goalShotEval->otherEval(myWS, oplayer.pos));

  if (goalGrade.is_better_than(Grade::BAD)) {
    Vector2d dir(myWS->otherGoal.pos-myWS->ball.pos);
    // [C]
    positions[players++] = (1.5*PLAYER_SIZE+FREE_KICK_FORBIDDEN_AREA)/norm(dir)*dir+
      myWS->ball.pos;
    DECISIONTRACE("goalshot possible");
  }

#ifdef USEDRAWTRACE
  {
    ostrstream text;
    text << "GoalShotEval Grade: " << goalGrade << (char)0;
    DRAWTRACE( TEXT(text.str(), Vector2d(-FIELD_LENGTH/2,-FIELD_WIDTH/2)) );
  }
#endif

  // calculate dangerous passes
  // dont look at players in free kick circle
  Grade passGrades[11];
  passGrades[num-1].set_IMPOSSIBLE();
  dist = sqr(FREE_KICK_FORBIDDEN_AREA-2);

  for (int pnum = 1; pnum < 12; pnum++)
    if ((pnum != num) &&
	(sqrnorm(myWS->ball.pos - myWS->otherTeam(pnum).pos) > dist)) {//[C]
      passGrades[pnum-1] = passEval->otherEval(myWS, num, pnum);
      // if this player can do a goalshot upgrade the pass
      if (goalShotEval->otherEval(myWS, myWS->otherTeam(pnum).pos).
	  is_better_than(Grade::AVERAGE))
	passGrades[pnum-1].upgrade();

      DRAWTRACE( LINE(myWS->ball.pos, myWS->otherTeam(pnum).pos) );
#ifdef USEDRAWTRACE
      {
	ostrstream text;
	text << passGrades[pnum-1] << (char)0;
	DRAWTRACE( TEXT(text.str(), myWS->otherTeam(pnum).pos) );
      }
#endif
    }
    else
      continue;

  // sort players according to their passgrades
  int dangerousPlayers[11], pnum(0);
  Grade g(Grade::PERFECT);
  while (g.is_better_than(Grade::BAD)) {
    for (int i = 0; i < 11; i++) {
      if (passGrades[i].is_like(g)) {
	dangerousPlayers[pnum++] = i+1;
	DECISIONTRACE(" player " << i+1 << " is dangerous with grade " << g);
      }
    }
    g.downgrade();
  }

  // calculate positions
  // dont look at players in free kick circle
  for (int i = 0; i < pnum; i++) {
    const WorldPlayer* player(&myWS->otherTeam(dangerousPlayers[i]));
    Vector2d dir(player->pos - myWS->ball.pos);
    if (sqrnorm(dir) <=  dist)
      continue;             //[C]
    positions[players++] = (1.5*PLAYER_SIZE+FREE_KICK_FORBIDDEN_AREA)/norm(dir)*dir
      + myWS->ball.pos;
    // if there is allready a player at this pos dont use it
    double d = sqrnorm(worldSnap->my.pos-positions[players-1]);
    for (int j = 1; j < 12; j++) {
      if (j == worldSnap->my.number) continue;
      if ( d < sqr(10) &&  //[C]*
	  sqrnorm(positions[players-1]-worldSnap->ourTeam(j).pos) < 
	  sqr(2*PLAYER_SIZE)) {
	players--;
	break;
	DECISIONTRACE( "not using pos " << positions[players] << 
		       " as used by " << j);
      }
	}
  }

  // min distance for players on the circle ??? agent.conf
  dist = 4.5;  //[C]
  double ang(dist/FREE_KICK_FORBIDDEN_AREA);

  OTHERTRACE("will remove positions on circle closer than: " << RAD_TO_DEG(ang) << " deg");

  // now check for positions which are close 

  // first sort according to angle around ball.pos 
  
  // index for accessing the positions
  int posv[11] = {0,1,2,3,4,5,6,7,8,9,10};

  // * Bubblesort *
  int swapIndex(players-1);
  do {
    int ulimit = swapIndex;
    swapIndex = 0;
    for (int i = 0; i < ulimit; i++) {
      if ((positions[posv[i+1]]-myWS->ball.pos).getAngle() < 
	  (positions[posv[i]]-myWS->ball.pos).getAngle())
	{
	  int h(posv[i]);
	  posv[i] = posv[i+1];
	  posv[i+1] = h;
	  swapIndex = i;
	}
    }
  } while(swapIndex > 0);

  CMDTRACE(ost << "Pos Angles: ";
           for (int i = 0; i < players; i++)
	   ost << RAD_TO_DEG((positions[posv[i]]-myWS->ball.pos).getAngle())
	       << endl; 
           );

  int newPos[11], last(0);
  newPos[0] = posv[0];
  pnum = (players > 0) ? 1 : 0;
  // now calc which positions are left
  for (int i = 0; i < players-1; i++) {
    if (fabs((positions[posv[last]]-myWS->ball.pos).getAngle()-
	(positions[posv[i+1]]-myWS->ball.pos).getAngle())
	< ang) {
      OTHERTRACE("removing pos with angle: " 
		 << RAD_TO_DEG((positions[posv[i+1]]-myWS->ball.pos).getAngle()));
      continue;
    }
    else {
      newPos[pnum++] = posv[i+1];
      last = i+1;
    }
  }

  last=0;
  Vector2d newPositions[11];
  for (int j = 0; j < players; j++)
    for (int i = 0; i < pnum; i++)
      if (positions[j] == positions[newPos[i]]) {
	newPositions[last++] = positions[j];
	break;
      }
  players = pnum;

  for (int i = 0; i < players; i++)
    positions[i] = newPositions[i];

#ifdef USEDRAWTRACE
  for (int i = 0; i < players; i++)
    DRAWTRACE( POSITION(positions[i]) );
#endif

  if (players == 0)
    DECISIONTRACE("No passes better than BAD found");

  return players;
}

double
StandardSituationModule::
calcUsableStamina( const double &factor,
		   const Snapshot* ws )
{
  double stamina(0);
  // if we have enough stamina run faster
  if (ws->senseBody.stamina > 
      factor*ws->my.staminaIncMax+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = factor*ws->my.staminaIncMax;

  return stamina;
}
